import os
import dgl
import torch as th
import dgl.sampling
from torch.utils.data import IterableDataset
from dgl.sampling.neighbor import select_topk
from dgl.data.utils import load_graphs, save_graphs


class SkipGramBatchSampler(IterableDataset):
    def __init__(self, hg, batch_size, window_size, rw_len=None):
        self.hg = hg
        self.g = dgl.to_homogeneous(hg).to('cpu')
        self.NID = self.g.ndata[dgl.NID]
        self.NTYPE = self.g.ndata[dgl.NTYPE]
        num_nodes = {}
        for i in range(th.max(self.NTYPE) + 1):
            num_nodes[self.hg.ntypes[i]] = int((self.NTYPE == i).sum())
        self.num_nodes = num_nodes
        self.num_ntypes = len(self.num_nodes)
        # self.weights = {
        #     etype: hg.in_degrees(etype=etype).float() ** 0.75
        #     for _, etype, _ in hg.canonical_etypes
        # }
        self.batch_size = batch_size
        self.window_size = window_size
        if rw_len is not None:
            self.rw_len = rw_len
        else:
            self.rw_len = self.window_size * 2
        self.neg_hetero = True
        self.edge_dict = {}
        self.ntypes = hg.ntypes

    def __iter__(self):
        '''         u = []
                    v = []
                    for i in range(self.window_size):
                        a = traces[:self.window_size*2-i]
                        b = traces[i:]
                        u.append(a)
                        v.append(b)'''

        # random select heads
        # select tails through random walk skgram
        while True:
            heads = th.randint(0, self.g.number_of_nodes(), (self.batch_size,))
            traces, _ = dgl.sampling.random_walk(self.g, heads, length=self.rw_len)
            heads, tails = self.traces2pos(traces, self.window_size)

            heads = (self.NID[heads], self.NTYPE[heads])
            tails = (self.NID[tails], self.NTYPE[tails])
            import copy
            neg_tails = copy.deepcopy(tails)
            for i in range(self.num_ntypes):
                mask = (neg_tails[1] == i)
                ntype = self.hg.ntypes[i]
                neg_tails[0][mask] = th.randint(0, self.hg.number_of_nodes(ntype), size=neg_tails[0][mask].shape)
            yield heads, tails, neg_tails

    def pre_process(self):
        heads = th.arange(self.g.number_of_nodes())
        traces, _ = dgl.sampling.random_walk(self.g, heads, length=self.window_size)
        heads, tails = self.traces2pos(traces, self.window_size)

        heads = (self.NID[heads], self.NTYPE[heads])
        tails = (self.NID[tails], self.NTYPE[tails])
        for i in range(self.num_ntypes):
            for j in range(self.num_ntypes):
                mask_h = (heads[1] == i)
                mask_t = (tails[1] == j)
                edge = (self.ntypes[i], self.ntypes[i] + '-' + self.ntypes[j], self.ntypes[j])
                self.edge_dict[edge] = (heads[0][mask_h], tails[0][mask_t])

    def traces2pos(self, traces, window_size):
        '''
        sample positive edges through skip gram
        '''
        idx_list_u = []
        idx_list_v = []
        batch_size = traces.shape[0]
        for b in range(batch_size):
            walk = traces[b]
            if -1 in walk:
                #walk = traces[b]
                mask = (walk != -1)
                walk = walk[mask]
            for i in range(len(walk)):
                for j in range(i - window_size, i):
                    if j >= 0:
                        idx_list_u.append(walk[j])
                        idx_list_v.append(walk[i])
                for j in range(i + 1, i + 1 + window_size):
                    if j < len(walk):
                        idx_list_u.append(walk[j])
                        idx_list_v.append(walk[i])
                    else:
                        break

        # [num_pos * batch_size]
        u = th.LongTensor(idx_list_u)
        v = th.LongTensor(idx_list_v)

        return v, u


class NeighborSampler(object):
    def __init__(self, hg, ntypes, num_nodes, device):
        # the new graph
        self.hg = hg
        self.ntypes = ntypes
        self.num_nodes = num_nodes
        self.sampler = dgl.dataloading.MultiLayerFullNeighborSampler(1)
        self.device = device

    def build_hetero_graph(self, heads, tails):
        edge_dict = {}
        num_ntypes = len(self.ntypes)
        for i in range(num_ntypes):
            for j in range(num_ntypes):
                edge = (self.ntypes[i], self.ntypes[i]+ '-' + self.ntypes[j], self.ntypes[j])
                mask = (heads[1] == i) & (tails[1] == j)
                edge_dict[edge] = (heads[0][mask], tails[0][mask])
        hg = dgl.heterograph(edge_dict, self.num_nodes)
        return hg

    def sample_from_item_pairs(self, heads, tails, neg_tails):
        # Create a graph with positive connections only and another graph with negative
        # connections only.
        pos_graph = self.build_hetero_graph(heads, tails)
        neg_graph = self.build_hetero_graph(heads, neg_tails)

        pos_graph, neg_graph = dgl.compact_graphs([pos_graph, neg_graph])
        pos_nodes = pos_graph.ndata[dgl.NID]
        seed_nodes = pos_nodes # same with neg_nodes from neg_graph

        blocks = self.sampler.sample_blocks(
            self.hg, seed_nodes, exclude_eids=None)
        return pos_graph, neg_graph, blocks


def assign_simple_node_features(ndata, g, ntypes, assign_id=False):
    """
    Copies data to the given block from the corresponding nodes in the original graph.
    """
    for ntype in ntypes:
        for col in g.nodes[ntype].data.keys():
            if not assign_id and col == dgl.NID:
                continue
            induced_nodes = ndata[dgl.NID][ntype]
            ndata[col] = {ntype: g.nodes[ntype].data[col][induced_nodes]}


def assign_features_to_blocks(blocks, g, ntypes):
    # For the first block (which is closest to the input), copy the features from
    # the original graph as well as the texts.
    # for ntype in ntypes:
    #     for col in g.nodes[ntype].data.keys():
    #         if not assign_id and col == dgl.NID:
    #             continue
    #         induced_nodes = blocks[0].srcnodes[ntype].data[dgl.NID]
    #         blocks[0].srcnodes[ntype].data[col] = g.nodes[ntype].data[col][induced_nodes]
    assign_simple_node_features(blocks[0].srcdata, g, ntypes)
    #assign_simple_node_features(blocks[-1].dstdata, g, ntypes)


class HetGNNCollator(object):
    def __init__(self, sampler, g):
        self.sampler = sampler
        self.g = g

    def collate_train(self, batches):
        heads, tails, neg_tails = batches[0]
        # Construct multilayer neighborhood via PinSAGE...
        pos_graph, neg_graph, blocks = self.sampler.sample_from_item_pairs(heads, tails, neg_tails)
        assign_features_to_blocks(blocks, self.g, self.g.ntypes)

        return pos_graph, neg_graph, blocks

    # def collate_test(self, samples):
    #     batch = th.LongTensor(samples)
    #     blocks = self.sampler.sample_blocks(batch)
    #     assign_features_to_blocks(blocks, self.g, self.g.ntypes)
    #     return blocks
    
    
class hetgnn_graph():
    """

    """
    def __init__(self, hg, n_dataset):
        self.hg = hg
        self.g = dgl.to_homogeneous(hg).to('cpu')
        self.NID = self.g.ndata[dgl.NID]
        self.NTYPE = self.g.ndata[dgl.NTYPE]
        num_nodes = {}
        for i in range(th.max(self.NTYPE) + 1):
            num_nodes[self.hg.ntypes[i]] = int((self.NTYPE == i).sum())
        self.num_nodes = num_nodes
        self.weight_column = 'w'
        self.n_dataset = n_dataset

    def get_hetgnn_graph(self, length, walks, restart_prob):
        fname = './openhgnn/output/HetGNN/{}_{}_{}.bin'.format(
            self.n_dataset, length, walks)
        if os.path.exists(fname):
            g, _ = load_graphs(fname)
            return g[0]
        else:
            g = self.build_hetgnn_graph(length, walks, restart_prob)
            save_graphs(fname, g)
            return g

    def build_hetgnn_graph(self, length, walks, restart_prob):
        #edges = [[[[],[]]] * len(self.num_nodes)] * len(self.num_nodes)
        edges = [[[[],[]], [[],[]], [[],[]]],
                 [[[],[]], [[],[]], [[],[]]],
                 [[[],[]], [[],[]], [[],[]]]]

        for i in range(self.g.number_of_nodes()):
            nodes = th.tensor([i]).repeat(walks)
            traces, types = dgl.sampling.random_walk(self.g, nodes, length=length, restart_prob=restart_prob)
            concat_vids, _, _, _ = dgl.sampling.pack_traces(traces, types)
            concat_types = th.index_select(self.NTYPE, 0, concat_vids)
            uid = concat_vids[0]
            utype = concat_types[0]
            for (vid, vtype) in zip(concat_vids, concat_types):
                edges[int(utype)][int(vtype)][0].append(self.NID[uid])
                edges[int(utype)][int(vtype)][1].append(self.NID[vid])

        edge_dict = {}
        k = {}
        num_ntypes = self.NTYPE.max() + 1
        for i in range(num_ntypes):
            for j in range(num_ntypes):
                edge = (self.hg.ntypes[j], self.hg.ntypes[j]+'-'+self.hg.ntypes[i], self.hg.ntypes[i])
                edge_dict[edge] = (th.tensor(edges[i][j][1]), th.tensor(edges[i][j][0]))
                if j == 2:
                    k[edge] = 3
                else:
                    k[edge] = 10

        neighbor_graph = dgl.heterograph(
            edge_dict,
            self.num_nodes
        )

        neighbor_graph = dgl.to_simple(neighbor_graph, return_counts=self.weight_column)
        counts = neighbor_graph.edata[self.weight_column]
        neighbor_graph = select_topk(neighbor_graph, k, self.weight_column)

        return neighbor_graph


import copy

from dgl._ffi.base import DGLError
from openhgnn import sampler

from torch.utils.data.dataloader import DataLoader
from torch.utils.data.sampler import BatchSampler
from torch.utils.data import TensorDataset
from openhgnn.trainerflow.base_flow import BaseFlow
import dgl
from networkx.algorithms.centrality.betweenness import edge_betweenness
import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from torch.autograd import Variable
import torch.nn.functional as F
from tqdm import tqdm, trange
from sklearn.model_selection import train_test_split
import random
import os
import pickle
import json
import time
from typing import List
import shutil
import copy
import pandas as pd
import math
from sklearn.metrics import (
    accuracy_score,
    auc,
    f1_score,
    mean_squared_error,
    precision_recall_curve,
    precision_recall_fscore_support,
    roc_auc_score,
    roc_curve,
)

from . import BaseFlow, register_flow
from ..tasks import build_task
from openhgnn.models import build_model
from openhgnn.models.SLiCE import SLiCE, SLiCEFinetuneLayer
from ..utils import extract_embed, EarlyStopping

from ..sampler.SLiCE_sampler import SLiCESampler
@register_flow("slicetrainer")
class SLiCETrainer(BaseFlow):
    def __init__(self,args):
        super(SLiCETrainer, self).__init__(args)
        self.out_dir=args.outdir
        self.pretrain_path=os.path.join(self.out_dir,'pretrain/')
        self.pretrain_save_path=os.path.join(self.pretrain_path,'best_pretrain_model.pt')
        self.finetune_path=os.path.join(self.out_dir,'finetune/')
        self.finetune_save_path=os.path.join(self.finetune_path,'best_finetune_model.pt')
        self.g=self.task.dataset.g
        self.g=dgl.to_homogeneous(self.g,edata=['train_mask','valid_mask','test_mask','label'])

        self.model=dict()
        self.model['pretrain']=SLiCE.build_model_from_args(self.args,self.g)
        self.model['finetune']=SLiCEFinetuneLayer.build_model_from_args(args)
        #loss function
        self.loss_fn=torch.nn.CrossEntropyLoss()
        #optimizer
        self.optimizer=dict()
        self.optimizer['pretrain']=optim.Adam(self.model['pretrain'].parameters(), lr=args.lr)
        self.optimizer['finetune']=optim.Adam(self.model['finetune'].parameters(), lr=args.ft_lr)

        self.patience=5 #for early stopping
        #number of epochs
        self.n_epochs=dict()
        self.n_epochs['pretrain']=args.n_epochs
        self.n_epochs['finetune']=args.ft_n_epochs
        #batch size
        self.batch_size=dict()
        self.batch_size['pretrain']=args.batch_size
        self.batch_size['finetune']=args.ft_batch_size

        self.labels = self.g.edata['label']
        self.idx=dict()
        #pretrain
        self.node_subgraphs=dict()
        #finetune
        self.edges=dict()
        self.edges_label=dict()

        self.graphs=dict()
        self.best_epoch=dict()
        self.is_pretrained=False
        self.is_finetuned=False
        self.threshold=None
    def preprocess(self):
        if not os.path.exists(self.pretrain_path):
            os.makedirs(self.pretrain_path)
        if not os.path.exists(self.finetune_path):
            os.makedirs(self.finetune_path)
        for task in ['train','valid','test']:
            #make directories
            if not os.path.exists(os.path.join(self.finetune_path,task)):
                os.makedirs(os.path.join(self.finetune_path,task))
            mask=self.g.edata[task+'_mask']
            index = torch.nonzero(mask.squeeze()).squeeze()
            if task=='train':
                self.idx['train']=index
            elif task=='valid':
                self.idx['valid']=index
            else:
                self.idx['test']=index
            self.edges[task]=self.g.find_edges(index)
            #finally, g should be a graph containing just train_edges
            self.graphs[task]=dgl.edge_subgraph(self.g,index)
        #sample walks
        self.sampler=SLiCESampler(self.g,self.graphs['train'],num_walks_per_node=self.args.n_pred,beam_width=self.args.beam_width,
                            max_num_edges=self.args.max_length,walk_type=self.args.walk_type,
                            path_option=self.args.path_option,save_path=self.out_dir)#full graph
        sampler=self.sampler
        #get dataloader for pretrain and finetune evaluation on link prediction
        g=self.g
        #pretrain
        node_walk_path=self.pretrain_path+'node_walks.bin'
        if os.path.exists(node_walk_path):
            node_walks,_=dgl.load_graphs(node_walk_path)
        else:
            node_walks=sampler.get_node_subgraph(g.nodes())
            dgl.save_graphs(node_walk_path,node_walks)
        random.shuffle(node_walks)
        total_len=len(node_walks)
        train_size=int(0.8*total_len)
        valid_size=int(0.1*total_len)
        self.node_subgraphs['train']=node_walks[:train_size]
        self.node_subgraphs['valid']=node_walks[train_size:train_size+valid_size]
        self.node_subgraphs['test']=node_walks[train_size+valid_size:]
        #finetune
        src,dst=g.find_edges(self.idx['train'])
        self.edges['train']=list(zip(src.tolist(),dst.tolist()))
        src,dst=g.find_edges(self.idx['valid'])
        self.edges['valid']=list(zip(src.tolist(),dst.tolist()))
        src,dst=g.find_edges(self.idx['test'])
        self.edges['test']=list(zip(src.tolist(),dst.tolist()))
        self.edges_label['train']=list()
        self.edges_label['valid']=[int(self.labels[x]) for x in self.idx['valid']]
        self.edges_label['test']=[int(self.labels[x]) for x in self.idx['test']]
        edges_label=self.edges_label
        #generate pretrain subgraph
        train_file=os.path.join(self.finetune_path,'train_edges.pickle')
        if os.path.exists(train_file):
            with open(train_file,'rb') as f:
                self.edges['train'],edges_label['train']=pickle.load(f)
        else:
            self.edges['train'],edges_label['train']=sampler.generate_false_edges2(self.edges['train'],train_file)
        self.edges['valid'],self.edges_label['valid']=sampler.shuffle_edge_label(self.edges['valid'],self.edges_label['valid'])
        self.edges['test'],self.edges_label['test']=sampler.shuffle_edge_label(self.edges['test'],self.edges_label['test'])
        #generate finetune subgraph
        for task in ['train','valid','test']:
            edges=self.edges[task]
            batch_size=self.batch_size['finetune']
            n_batch=int(len(edges)/batch_size)
            total_len=len(edges)
            for batch in range(n_batch):
                i=batch*batch_size
                if i+batch_size<total_len:
                    end=i+batch_size
                else:
                    end=total_len
                batch_file=os.path.join(self.finetune_path,'{}/edge_subgraph_{}.bin'.format(task,batch))
                #pair_file=self.finetune_path+'{}/pair_subgraph_{}.pickle'.format(task,batch)
                if not os.path.exists(batch_file):
                    subgraph_list=self.sampler.get_edge_subgraph(self.edges[task][i:end])
                    dgl.save_graphs(batch_file,subgraph_list)

    def train(self):
        self.preprocess()
        self.pretrain()
        self.finetune()
    def pretrain(self):
        print("Start Pretraining...")
        stopper=EarlyStopping(self.patience)
        batch_size=self.batch_size['pretrain']
        self.model['pretrain'].train()
        self.is_pretrained=True
        if os.path.exists(self.pretrain_save_path):
            pass
        for epoch in range(self.n_epochs['pretrain']):
            print("Epoch {}:".format(epoch))
            i=0
            total_len=len(self.node_subgraphs['train'])
            n_batch=math.ceil(total_len/batch_size)
            bar=tqdm(range(n_batch))
            avg_loss=0
            for batch in bar:
                i=batch*batch_size
                if i+batch_size<total_len:
                    subgraph_list=self.node_subgraphs['train'][i:i+batch_size]
                else:
                    subgraph_list=self.node_subgraphs['train'][i:]
                pred_data,true_data=self.model['pretrain'](subgraph_list)
                loss=self.loss_fn(pred_data.transpose(1,2).cuda(),true_data.cuda())
                avg_loss+=float(loss)
                self.optimizer['pretrain'].zero_grad()
                loss.backward()
                self.optimizer['pretrain'].step()
                i+=batch_size
                bar.set_description("Batch {} Loss: {:.3f}".format(batch,loss))
            #torch.save(self.model['pretrain'],self.pretrain_path+'model_'+str(ii)+'SLiCE.pt')
            avg_loss=avg_loss/n_batch
            print("AvgLoss: {:.3f}".format(avg_loss))
            early_stop=stopper.loss_step(avg_loss,self.model['pretrain'])
            if early_stop:
                print('Early Stop!\tEpoch:' + str(epoch))
                break
        
        self.best_epoch['pretrain']=epoch
        torch.save(self.model['pretrain'].state_dict(),self.pretrain_save_path)
        print("Evaluating for pretraining...")
        
    def finetune(self):
        if not os.path.exists(self.pretrain_save_path):
            print("Model not pretrained!")
        else:
            ck_pt=torch.load(self.pretrain_save_path)
        self.model['pretrain'].load_state_dict(ck_pt)
        self.model['pretrain'].eval()
        self.model['pretrain'].set_fine_tuning()
        self.model['finetune'].train()
        print("Start Finetuning...")
        stopper=EarlyStopping(self.patience)
        batch_size=self.batch_size['finetune']
        for epoch in range(self.n_epochs['finetune']):
            batch=0
            total_len=len(self.edges['train'])
            print("Eopch {}:".format(epoch))
            n_batch=math.ceil(total_len/batch_size)
            bar=tqdm(range(n_batch))
            avg_loss=0
            for batch in bar:
                i=batch*batch_size
                if i+batch_size<total_len:
                    end=i+batch_size
                else:
                    end=total_len
                batch_file=os.path.join(self.finetune_path,'{}/edge_subgraph_{}.bin'.format('train',batch))
                
                if os.path.exists(batch_file):
                    subgraph_list,_=dgl.load_graphs(batch_file)
                else:
                    subgraph_list=self.sampler.get_edge_subgraph(self.edges['train'][i:end])
                    dgl.save_graphs(batch_file,subgraph_list)
                self.model['pretrain'].set_fine_tuning()

                with torch.no_grad():
                    _,layer_output,_=self.model['pretrain'](subgraph_list)
                pred_scores,_,_=self.model['finetune'](layer_output)
                loss=F.binary_cross_entropy(pred_scores,torch.tensor(self.edges_label['train'][i:end],dtype=torch.float).reshape(-1,1).cuda())
                bar.set_description('Batch {}: Loss:{:.3f}'.format(batch,loss))
                avg_loss+=float(loss)
                self.optimizer['finetune'].zero_grad()
                loss.backward()
                self.optimizer['finetune'].step()
            torch.save(self.model['finetune'].state_dict(),self.finetune_path+'model_'+str(epoch)+'SLiCE.pt')
            avg_loss=avg_loss/n_batch
            print("AvgLoss: {:.3f}".format(avg_loss))
            early_stop=stopper.loss_step(avg_loss,self.model['finetune'])
            if early_stop:
                print('Early Stop!\tEpoch:' + str(epoch))
                break
        self.model['finetune']=stopper.best_model
        torch.save(self.model,self.finetune_save_path)
        #run validation to find the best epoch
        self.is_finetuned=True
        print("Evaluating for pretraining...")
        self._test_step()
    def _test_step(self):
        with torch.no_grad():
            #validation and find best threshold
            pred_data={'train':[],'valid':[],'test':[]}
            true_data={'train':[],'valid':[],'test':[]}
            self.model['pretrain'].eval()
            for task in ['valid','test']:
                
                total_len=len(self.edges[task])
                batch_size=self.batch_size['finetune']
                n_batch=int(total_len/batch_size)
                for batch in range(n_batch):
                    i=batch*batch_size
                    if i+batch_size<total_len:
                        end=i+batch_size
                    else:
                        end=total_len
                    #get edge subgraphs for test
                    batch_file=os.path.join(self.finetune_path,'{}/edge_subgraph_{}.bin'.format(task,batch))
                    if os.path.exists(batch_file):
                        subgraph_list,_=dgl.load_graphs(batch_file)
                    else:
                        subgraph_list=self.sampler.get_edge_subgraph(self.edges[task][i:end])
                        dgl.save_graphs(batch_file,subgraph_list)
                    #get score and label
                    #output: 100*7*200  layer_output: 100*6*7*200
                    output,layer_output,_=self.model['pretrain'](subgraph_list)
                    if not self.is_finetuned:
                        source_embed = output[:, 0, :].unsqueeze(1)
                        target_embed = output[:, 1, :].unsqueeze(1).transpose(1, 2)
                        score = torch.bmm(source_embed, target_embed).squeeze(1)#embedding相乘得到相似度分数
                        score = torch.sigmoid(score).data.cpu().numpy().tolist()
                    else:
                        #score:[ft_batch_size,1]
                        #src_embedding/dst_embedding:[ft_batch_size,1,embedding_dim]
                        score,_,_=self.model['finetune'](layer_output)
                    labels=self.edges_label[task][i:end]
                    for ii, _ in enumerate(score):
                        pred_data[task].append(float(score[ii][0]))
                        true_data[task].append(labels[ii])
                    i+=batch_size
            #test and get result
            real_true_data=np.array(true_data['valid'],dtype=np.int)
            self.threshold=self.get_threshold(real_true_data,pred_data['valid'])[0]
            prediction_data=pred_data['test']
            sorted_pred = prediction_data[:]
            sorted_pred.sort()
            # threshold = sorted_pred[-true_num]
            y_pred = np.zeros(len(prediction_data), dtype=np.int32)

            for i, _ in enumerate(prediction_data):
                if prediction_data[i] >= self.threshold:
                    y_pred[i] = 1

            y_true = np.array(true_data['test'])
            y_scores = np.array(prediction_data)
            ps, rs, _ = precision_recall_curve(y_true, y_scores)
            if self.is_finetuned:
                header="Finetuning"
            else:
                header="Pretraining"
            print(f"----------------------Testing for {header}()------------")
            print(
                f"y_true.shape: {y_true.shape}, y_scores.shape: {y_scores.shape}"
                f", y_pred.shape: {y_pred.shape}"
            )
            try:
                roc_auc = roc_auc_score(y_true, y_scores)
            except ValueError:
                roc_auc = 'UNDEFINED'
            f1 = f1_score(y_true, y_pred)
            auc_value = auc(rs,ps)
            
            print(
                f"{header} : ROC-AUC: {roc_auc},"
                f" F1: {f1}, AUC: {auc_value}"
            )
    def get_threshold(self, target, predicted): 
        fpr, tpr, threshold = roc_curve(target, predicted,pos_label=1)
        i = np.arange(len(tpr),dtype=np.int)
        roc = pd.DataFrame(
            { 
                "tf": pd.Series(tpr - (1 - fpr), index=i),
                "threshold": pd.Series(threshold, index=i),
            }
        )
        print()
        roc_t = roc.loc[(roc.tf - 0).abs().argsort()[:1]]
        return list(roc_t["threshold"])
    def loss_calculation(self, pos_score, neg_score):
        # an example hinge loss
        loss = []
        for i in pos_score:
            loss.append(F.logsigmoid(pos_score[i]))
            loss.append(F.logsigmoid(-neg_score[i]))
        loss = torch.cat(loss)
        return -loss.mean()

    def ScorePredictor(self, edge_subgraphs, pairs, x):
        #x:[batch_size*num_nodes*embed_dim]
        score=[]
        labels=[]
        for ii,edge_subgraph in enumerate(edge_subgraphs):
            src_embed=x[ii,0,:]
            dst_embed=x[ii,1,:]
            score.append(torch.dot(src_embed,dst_embed))
            src,dst,label=pairs[ii]
            labels.append(label)
        score=torch.sigmoid(torch.tensor(score))
        res=F.binary_cross_entropy(score,torch.FloatTensor(labels))
        return res
    def nid_to_id(self,subgraph,src):
        for ii,each in enumerate(subgraph.ndata[dgl.NID]):
            if each==src:
                return ii
        return -1


import os
import dgl
import pickle
import random
import torch as th
import numpy as np
from scipy.sparse import coo_matrix
from dgl.dataloading.negative_sampler import _BaseNegativeSampler
from dgl import backend as F
from dgl.data.utils import load_graphs, save_graphs
import dgl.sampling
from collections import Counter
from torch.utils.data import IterableDataset, DataLoader


def get_node_data_loader(node_neighbors_min_num: int, n_layers: int,
                         graph: dgl.DGLGraph, batch_size: int, sampled_node_type: str,
                         train_idx: th.Tensor, valid_idx: th.Tensor, test_idx: th.Tensor,
                         shuffle: bool = True, drop_last: bool = False, num_workers: int = 4):
    """
    get graph node data loader, including train_loader, val_loader and test_loader
    :return:
    """
    # list of neighbors to sample per edge type for each GNN layer
    sample_nodes_num = []
    for layer in range(n_layers):
        sample_nodes_num.append({etype: node_neighbors_min_num + layer for etype in graph.canonical_etypes})

    # neighbor sampler
    sampler = dgl.dataloading.MultiLayerNeighborSampler(sample_nodes_num)

    train_loader = dgl.dataloading.NodeDataLoader(
        graph, {sampled_node_type: train_idx}, sampler,
        batch_size=batch_size, shuffle=shuffle, drop_last=drop_last, num_workers=num_workers)

    val_loader = dgl.dataloading.NodeDataLoader(
        graph, {sampled_node_type: valid_idx}, sampler,
        batch_size=batch_size, shuffle=shuffle, drop_last=drop_last, num_workers=num_workers)

    test_loader = dgl.dataloading.NodeDataLoader(
        graph, {sampled_node_type: test_idx}, sampler,
        batch_size=batch_size, shuffle=shuffle, drop_last=drop_last, num_workers=num_workers)

    return train_loader, val_loader, test_loader

class MP2vecCollator(object):
    def __init__(self, ntypes, num_nodes):
        # the new graph
        self.ntypes = ntypes
        self.num_nodes = num_nodes

    def build_hetero_graph(self, heads, tails):
        edge_dict = {}
        num_ntypes = len(self.ntypes)
        for i in range(num_ntypes):
            for j in range(num_ntypes):
                edge = (self.ntypes[i], self.ntypes[i] + '-' + self.ntypes[j], self.ntypes[j])
                mask = (heads[1] == i) & (tails[1] == j)
                edge_dict[edge] = (heads[0][mask], tails[0][mask])
        hg = dgl.heterograph(edge_dict, self.num_nodes)
        return hg

    def construct_from_item_pairs(self, heads, tails, neg_tails):
        pos_graph = self.build_hetero_graph(heads, tails)
        neg_graph = self.build_hetero_graph(heads, neg_tails)

        pos_graph, neg_graph = dgl.compact_graphs([pos_graph, neg_graph])
        # pos_nodes = pos_graph.ndata[dgl.NID]
        # seed_nodes = pos_nodes  # same with neg_nodes from neg_graph

        return pos_graph, neg_graph, [pos_graph]

    def collate_train(self, batches):
        heads, tails, neg_tails = batches[0]
        pos_graph, neg_graph, block = self.construct_from_item_pairs(heads, tails, neg_tails)
        return pos_graph, neg_graph, block



class randomwalk_on_heterograph(object):
    def __init__(self, hg):
        self.hg = hg
        self.g = dgl.to_homogeneous(hg).to('cpu')
        self.NID = self.g.ndata[dgl.NID]
        self.NTYPE = self.g.ndata[dgl.NTYPE]
        num_nodes = {}
        for i in range(th.max(self.NTYPE) + 1):
            num_nodes[self.hg.ntypes[i]] = int((self.NTYPE == i).sum())
        self.num_nodes = num_nodes
        self.weight_column = 'w'

    def randomwalk(self, length, walks, restart_prob):
        from dgl.sampling.neighbor import select_topk
        from dgl.sampling.pinsage import RandomWalkNeighborSampler
        #path, _ = dgl.sampling.random_walk(self.g, nodes, length=length)
        #traces, types = dgl.sampling.random_walk(self.g, nodes, length=length)
        edges = [[[[],[]], [[],[]], [[],[]]],
                 [[[],[]], [[],[]], [[],[]]],
                 [[[],[]], [[],[]], [[],[]]]]
        for i in range(self.g.number_of_nodes()):
            nodes = th.tensor([i]).repeat(walks)
            traces, types = dgl.sampling.random_walk(self.g, nodes, length=length, restart_prob=restart_prob)
            concat_vids, _, _, _ = dgl.sampling.pack_traces(traces, types)
            concat_types = th.index_select(self.NTYPE, 0, concat_vids)
            uid = concat_vids[0]
            utype = concat_types[0]
            for (vid, vtype) in zip(concat_vids, concat_types):
                # if vid == uid:
                #     pass
                # else:
                edges[int(utype)][int(vtype)][0].append(self.NID[uid])
                edges[int(utype)][int(vtype)][1].append(self.NID[vid])

        from dgl import convert
        from dgl import transform
        edge_dict = {}
        k = {}
        num_ntypes = self.NTYPE.max() + 1
        for i in range(num_ntypes):
            for j in range(num_ntypes):
                edge = (self.hg.ntypes[j], self.hg.ntypes[j]+'-'+self.hg.ntypes[i], self.hg.ntypes[i])
                edge_dict[edge] = (th.tensor(edges[i][j][1]), th.tensor(edges[i][j][0]))
                if j == 2:
                    k[edge] = 3
                else:
                    k[edge] = 10

        neighbor_graph = convert.heterograph(
            edge_dict,
            self.num_nodes
        )

        neighbor_graph = transform.to_simple(neighbor_graph, return_counts=self.weight_column)
        counts = neighbor_graph.edata[self.weight_column]
        neighbor_graph = select_topk(neighbor_graph, k, self.weight_column)
        from dgl.data.utils import load_graphs, save_graphs
        fname = './openhgnn/output/HetGNN/Academic.bin'
        save_graphs(fname, neighbor_graph)
        return

    def randomwalk_with_restart(self, nodes, length, restart_prob=None):

        path = dgl.sampling.random_walk(self.g, nodes, length=length, restart_prob=restart_prob)
        return path

    def full_rwr(self, walks, length, restart_prob):
        n = self.g.number_of_nodes()
        # expand to 'walks' times n
        nodes = th.arange(n).repeat(walks)
        path = self.randomwalk(nodes, length)
        return path

    def path_hetro(self):
        pass

    def neighbour_hetero(self, path):
        edges = {}
        for p in path[0]:
            center_node = p[0]
            neigh_nodes = p[1:]
            c_nid = int(self.NID[center_node])
            c_ntype = int(self.NTYPE[center_node])
            edges.setdefault(c_ntype, {})
            edges[c_ntype].setdefault(c_nid, {})
            for n in neigh_nodes:
                if n >= 0:
                    n_nid = self.NID[n]
                    n_ntype = int(self.NTYPE[n])
                    edges[c_ntype][c_nid].setdefault(n_ntype, [])
                    edges[c_ntype][c_nid][n_ntype].append(n_nid)
                else:
                    break
        import dgl.sampling.select_topk
        a = 1
        # src = F.reshape(paths[:, self.metapath_hops::self.metapath_hops], (-1,))
        # dst = F.repeat(paths[:, 0], self.num_traversals, 0)
        #
        # src_mask = (src != -1)
        # src = F.boolean_mask(src, src_mask)
        # dst = F.boolean_mask(dst, src_mask)
        # import dgl.convert as convert
        # # count the number of visits and pick the K-most frequent neighbors for each node
        # neighbor_graph = convert.heterograph(
        #     {(self.ntype, '_E', self.ntype): (src, dst)},
        #     {self.ntype: self.G.number_of_nodes(self.ntype)}
        # )
        # neighbor_graph = transform.to_simple(neighbor_graph, return_counts=self.weight_column)
        # counts = neighbor_graph.edata[self.weight_column]
        #
        # neighbor_graph = select_topk(neighbor_graph, self.num_neighbors, self.weight_column)
        # selected_counts = F.gather_row(counts, neighbor_graph.edata[EID])
        # neighbor_graph.edata[self.weight_column] = selected_counts



class pro_sampler(_BaseNegativeSampler):

    def __init__(self, k, p):
        self.k = k
        self.p = p

    def _generate(self, g, eids, canonical_etype):
        _, _, vtype = canonical_etype
        shape = F.shape(eids)
        dtype = F.dtype(eids)
        ctx = F.context(eids)
        shape = (shape[0] * self.k,)
        src, _ = g.find_edges(eids, etype=canonical_etype)
        src = F.repeat(src, self.k, 0)
        dst = np.random.choice(np.arange(0, g.number_of_nodes()), shape, replace=True, p=self.p)
        # dst = F.randint(shape, dtype, ctx, 0, g.number_of_nodes(vtype))
        dst = th.tensor(dst, dtype=dtype, device=ctx)
        return src, dst


def gen_neg_edges(g, num_neg, device):
    if not g.is_homogeneous:
        g_homo = dgl.to_homogeneous(g)
    else:
        g_homo = g
    node_degrees = g_homo.out_degrees().to('cpu').numpy()
    node_weights = np.power(node_degrees, 0.75)
    node_probs = node_weights / np.sum(node_weights)

    # neg_sampler = dgl.dataloading.negative_sampler.Uniform(num_neg)
    # neg_edges = neg_sampler(g_homo, th.arange(0, g_homo.num_edges(), dtype=th.int64, device=device))
    neg_sampler = pro_sampler(num_neg, node_probs)
    neg_edges = neg_sampler(g_homo, th.arange(0, g_homo.num_edges(), dtype=th.int64, device=device))
    # tensors used as indices must be long, byte or bool tensors， so it should be tensor.int64
    return neg_edges


def _get_neg_edge(g, epoch_seed, n_dataset, num_neg, device):
    fname = './openhgnn/output/NSHE/{}_NE-rate={:.0f}_seed={}.dat'.format(
        n_dataset, num_neg, epoch_seed)
    if os.path.exists(fname):
        # load the existing negative edge sampling file
        with open(fname, 'rb') as f:
            try:
                epoch_data = pickle.load(f)
                neg_edges = epoch_data['neg_edges']
            except EOFError:
                os.remove(fname)
                print(epoch_seed, fname)
    else:
        # sample
        neg_edges = gen_neg_edges(g, num_neg=num_neg, device=device)
        # save
        data_to_save = {'neg_edges': neg_edges}
        with open(fname, 'wb') as f:
            pickle.dump(data_to_save, f, protocol=pickle.HIGHEST_PROTOCOL)
    return neg_edges

def gen_ns_instances(g, num_ns_neg):

    def _get_current_sample_types(prev_sampled_types, pos_ns):
        u_node_type = []
        v_node_type = []
        etypes = []
        for t in set(prev_sampled_types):
            for etype in g.canonical_etypes:
                if t in etype[0]:
                    if etype[2] not in pos_ns.keys():
                        u_node_type.append(etype[0])
                        v_node_type.append(etype[2])
                        etypes.append(etype[1])
        return u_node_type, etypes, v_node_type

    def _sample_pos_ns(i, target_t):
        pos_ns = {target_t: i}
        prev_sampled_types = [target_t]
        prev_sampled_types, etypes, current_sample_types = _get_current_sample_types(prev_sampled_types, pos_ns)
        prev_t = i
        while len(current_sample_types) > 0:
        #BFS
            for (u_ntype, etype, v_ntype) in zip(prev_sampled_types, etypes, current_sample_types):
                # find neighbors of type t
                prev_nid = pos_ns[u_ntype]
                t_neighbors = g.out_edges(prev_nid, etype=etype)[1]
                # random select one as postive ns_instance
                if len(t_neighbors) == 0:  # if there is no neighbor to select
                    print('Node {} has no {} type point!!'.format(id, v_ntype))
                    return None
                elif len(t_neighbors) == 1:  # if there is only one selection
                    r = 0
                elif len(t_neighbors) > 1:  # if there is only one selection
                    r = random.randint(0, len(t_neighbors) - 1)
                pos_ns[v_ntype] = t_neighbors[r]

            prev_sampled_types = current_sample_types
            prev_sampled_types, etypes, current_sample_types = _get_current_sample_types(prev_sampled_types, pos_ns)
        pos_ns['label'] = 1
        return pos_ns

    ns_ins_list = []
    for target_t in g.ntypes:
        # Sample using target_t
        target_nodes = g.nodes(target_t)
        num_target_node = g.num_nodes(target_t)
        ns_ins = []
        for i in target_nodes:
            # ================== pos ns instnace sampling ==================
            pos_ns = _sample_pos_ns(i, target_t)
            if pos_ns is not None:
                ns_ins.append(pos_ns)
            else:
                continue
            # ================== neg ns instnace sampling ==================
            for _ in range(num_ns_neg):
                neg_ns = pos_ns.copy()
                neg_node = target_nodes[random.randint(0, num_target_node - 1)]
                # replace type in schema instance
                neg_ns[target_t] = neg_node
                neg_ns['label'] = 0  # BCE_loss, negative samples label = 0
                ns_ins.append(neg_ns)
        ns_ins_dict = {}
        ns_ins_dict['target_type'] = target_t
        ns_ins_dict['label'] = th.tensor([x['label'] for x in ns_ins])
        ns_ins_dict[target_t] = th.tensor([x[target_t] for x in ns_ins])
        for ntype in g.ntypes:
            if ntype != target_t:
                ns_ins_dict[ntype] = th.tensor([x[ntype] for x in ns_ins])

        ns_ins_list.append(ns_ins_dict)
    return ns_ins_list


def _get_ns_instance(g, epoch_seed, n_dataset, num_ns_neg):
    fname = './openhgnn/output/NSHE/{}_NS-rate={:.0f}_seed={}.dat'.format(
        n_dataset, num_ns_neg, epoch_seed)
    if os.path.exists(fname):
        # load the existing negative network schemas sampling file
        with open(fname, 'rb') as f:
            try:
                epoch_data = pickle.load(f)
            except EOFError:
                print(epoch_seed, fname)
        ns_ins_list = epoch_data['ns_instances']
    else:

        ns_ins_list = gen_ns_instances(g, num_ns_neg)
        # save
        data_to_save = {
            'ns_instances': ns_ins_list}
        with open(fname, 'wb') as f:
            pickle.dump(data_to_save, f, protocol=pickle.HIGHEST_PROTOCOL)

    return ns_ins_list


def get_epoch_samples(g, epoch, dataset, ns_neg, device):
    """
    Renew ns_instances and neg_edges in every epoch:
    1. get the seed for current epoch
    2. find using seed
        Y: load the file
        N: sample again and save
    """

    epoch_seed = np.random.randint(1000)
    np.random.seed(epoch_seed)
    neg_edges = _get_neg_edge(g, epoch_seed, dataset, 1, device)
    ns_samples = _get_ns_instance(g, epoch_seed, dataset, ns_neg)

    return neg_edges, ns_samples





"""Heterograph NN modules"""
import torch as th
import torch.nn as nn

__all__ = ['HeteroGraphConv']

class HeteroGraphConv(nn.Module):
    r"""A generic module for computing convolution on heterogeneous graphs.

    The heterograph convolution applies sub-modules on their associating
    relation graphs, which reads the features from source nodes and writes the
    updated ones to destination nodes. If multiple relations have the same
    destination node types, their results are aggregated by the specified method.

    If the relation graph has no edge, the corresponding module will not be called.



    Parameters
    ----------
    mods : dict[str, nn.Module]
        Modules associated with every edge types. The forward function of each
        module must have a `DGLHeteroGraph` object as the first argument, and
        its second argument is either a tensor object representing the node
        features or a pair of tensor object representing the source and destination
        node features.
    aggregate : str, callable, optional
        Method for aggregating node features generated by different relations.
        Allowed string values are 'sum', 'max', 'min', 'mean', 'stack'.
        The 'stack' aggregation is performed along the second dimension, whose order
        is deterministic.
        User can also customize the aggregator by providing a callable instance.
        For example, aggregation by summation is equivalent to the follows:

        .. code::

            def my_agg_func(tensors, dsttype):
                # tensors: is a list of tensors to aggregate
                # dsttype: string name of the destination node type for which the
                #          aggregation is performed
                stacked = torch.stack(tensors, dim=0)
                return torch.sum(stacked, dim=0)

    Attributes
    ----------
    mods : dict[str, nn.Module]
        Modules associated with every edge types.
    """
    def __init__(self, mods):
        super(HeteroGraphConv, self).__init__()
        self.mods = nn.ModuleDict(mods)
        # Do not break if graph has 0-in-degree nodes.
        # Because there is no general rule to add self-loop for heterograph.
        for _, v in self.mods.items():
            set_allow_zero_in_degree_fn = getattr(v, 'set_allow_zero_in_degree', None)
            if callable(set_allow_zero_in_degree_fn):
                set_allow_zero_in_degree_fn(True)

    def forward(self, g, inputs, mod_args=None, mod_kwargs=None):
        """Forward computation

        Invoke the forward function with each module and aggregate their results.

        Parameters
        ----------
        g : DGLHeteroGraph
            Graph data.
        inputs : dict[str, Tensor] or pair of dict[str, Tensor]
            Input node features.
        mod_args : dict[str, tuple[any]], optional
            Extra positional arguments for the sub-modules.
        mod_kwargs : dict[str, dict[str, any]], optional
            Extra key-word arguments for the sub-modules.

        Returns
        -------
        dict[str, Tensor]
            Output representations for every types of nodes.
        """
        if mod_args is None:
            mod_args = {}
        if mod_kwargs is None:
            mod_kwargs = {}
        outputs = {nty : [] for nty in g.dsttypes}
        if isinstance(inputs, tuple) or g.is_block:
            if isinstance(inputs, tuple):
                src_inputs, dst_inputs = inputs
            else:
                src_inputs = inputs
                dst_inputs = {k: v[:g.number_of_dst_nodes(k)] for k, v in inputs.items()}

            for stype, etype, dtype in g.canonical_etypes:
                rel_graph = g[stype, etype, dtype]
                if rel_graph.number_of_edges() == 0:
                    continue
                if stype not in src_inputs or dtype not in dst_inputs:
                    continue
                dstdata = self.mods[etype](
                    rel_graph,
                    (src_inputs[stype], dst_inputs[dtype]),
                    *mod_args.get(etype, ()),
                    **mod_kwargs.get(etype, {}))
                outputs[dtype].append(dstdata)
        else:
            for stype, etype, dtype in g.canonical_etypes:
                rel_graph = g[stype, etype, dtype]
                if rel_graph.number_of_edges() == 0:
                    continue
                if stype not in inputs:
                    continue
                dstdata = self.mods[etype](
                    rel_graph,
                    (inputs[stype], inputs[dtype]),
                    *mod_args.get(etype, ()),
                    **mod_kwargs.get(etype, {}))
                outputs[dtype].append(dstdata)

        return outputs


import torch.nn as nn

act_dict = {
    'relu': nn.ReLU(),
    'relu6': nn.ReLU6(),
    'sigmoid': nn.Sigmoid(),
    'lrelu': nn.LeakyReLU(negative_slope=0.5),
    'tanh': nn.Tanh(),
    'elu': nn.ELU(),
    'prelu': nn.PReLU(),
    'selu': nn.SELU(),
    'lrelu_01': nn.LeakyReLU(negative_slope=0.1),
    'lrelu_025': nn.LeakyReLU(negative_slope=0.25),
    'lrelu_05': nn.LeakyReLU(negative_slope=0.5),
}



import dgl
import torch as th
import torch.nn as nn
import torch.nn.functional as F
from dgl.nn.pytorch import GraphConv, EdgeWeightNorm
from ..utils import transform_relation_graph_list
from . import BaseModel, register_model


@register_model('MHNF')
class MHNF(BaseModel):
    r"""

        Description
        -----------
        MHNF from paper `Multi-hop Heterogeneous Neighborhood information Fusion graph representation learning
        <https://arxiv.org/pdf/2106.09289.pdf>`__.

        Given a heterogeneous graph :math:`G` and its edge relation type set :math:`\mathcal{R}`.Then we can extract l-hops hybrid adjacency matrix list
        in HMAE model. The hybrid adjacency matrix list can be used in HLHIA model to generate l-hops representations. Then HSAF
        model use attention mechanism to aggregate l-hops representations and because of multi-channel conv, the
        HSAF model also  aggregates different channels l-hops representations to generate a final representation.
        You can see detail operation in correspond model.

        Parameters
        ----------
            num_edge_type : int
                Number of relations.
            num_channels : int
                Number of conv channels.
            in_dim : int
                The dimension of input feature.
            hidden_dim : int
                The dimension of hidden layer.
            num_class : int
                Number of classification type.
            num_layers : int
                Length of hybrid metapath.
            category : string
                Type of predicted nodes.
            norm : bool
                If True, the adjacency matrix will be normalized.
            identity : bool
                If True, the identity matrix will be added to relation matrix set.

    """
    @classmethod
    def build_model_from_args(cls, args, hg):
        if args.identity == True:
            num_edge_type = len(hg.canonical_etypes) + 1
        else:
            num_edge_type = len(hg.canonical_etypes)
        # add self-loop edge
        return cls(num_edge_type=num_edge_type, num_channels=args.num_channels,
                    in_dim=args.hidden_dim, hidden_dim=args.hidden_dim, num_class=args.out_dim,
                    num_layers=args.num_layers, category=args.category, norm=args.norm_emd_flag, identity=args.identity)

    def __init__(self, num_edge_type, num_channels, in_dim, hidden_dim, num_class, num_layers, category, norm, identity):
        super(MHNF, self).__init__()
        self.num_edge_type = num_edge_type
        self.num_channels = num_channels
        self.in_dim = in_dim
        self.hidden_dim = hidden_dim
        self.num_class = num_class
        self.num_layers = num_layers
        self.is_norm = norm
        self.category = category
        self.identity = identity

        self.HSAF = HSAF(num_edge_type, self.num_channels, self.num_layers, self.in_dim, self.hidden_dim)
        self.linear = nn.Linear(self.hidden_dim, self.num_class)
        self.category_idx = None
        self.A = None
        self.h = None

    def forward(self, hg, h=None):
        with hg.local_scope():
            #Ws = []
            hg.ndata['h'] = h
            # * =============== Extract edges in original graph ================
            if self.category_idx is None:
                self.A, h, self.category_idx = transform_relation_graph_list(hg, category=self.category,
                                                        identity=self.identity)
            else:
                g = dgl.to_homogeneous(hg, ndata='h')
                h = g.ndata['h']
            # g = dgl.to_homogeneous(hg, ndata='h')
            #X_ = self.gcn(g, self.h)
            A = self.A
            final_representation = self.HSAF(A, h)
            y = self.linear(final_representation)
            return {self.category: y[self.category_idx]}


class HSAF(nn.Module):
    r'''

        Description
        -----------
        HSAF: Hierarchical Semantic Attention Fusion

        The HSAF model use two level attention mechanism to generate final representation

        * Hop-level attention

          .. math::
              \alpha_{i, l}^{\Phi_{p}}=\sigma\left[\delta^{\Phi_{p}} \tanh \left(W^{\Phi_{p}} Z_{i, l}^{\Phi_{p}}\right)\right]

          In which, :math:`\alpha_{i, l}^{\Phi_{p}}` is the importance of the information :math:`\left(Z_{i, l}^{\Phi_{p}}\right)`
          of the l-th-hop neighbors of node i under the path :math:`\Phi_{p}`, and :math:`\delta^{\Phi_{p}}` represents the learnable matrix.

          Then normalize :math:`\alpha_{i, l}^{\Phi_{p}}`

          .. math::
              \beta_{i, l}^{\Phi_{p}}=\frac{\exp \left(\alpha_{i, l}^{\Phi_{p}}\right)}{\sum_{j=1}^{L} \exp \left(\alpha_{i, j}^{\Phi_{p}}\right)}

          Finally, we get hop-level attention representation in one hybrid metapath.

          .. math::
              Z_{i}^{\Phi_{p}}=\sum_{l=1}^{L} \beta_{l}^{\Phi_{p}} Z_{l}^{\Phi_{p}}

        * Channel-level attention

          It also can be seen as multi-head attention mechanism.

          .. math::
              \alpha_{i, \Phi_{p}}=\sigma\left[\delta \tanh \left(W Z_{i}^{\Phi_{p}}\right)\right.

          Then normalize :math:`\alpha_{i, \Phi_{p}}`

          .. math::
              \beta_{i, \Phi_{p}}=\frac{\exp \left(\alpha_{i, \Phi_{p}}\right)}{\sum_{p^{\prime} \in P} \exp \left(\alpha_{\Phi_{p^{\prime}}}\right)}

          Finally, we get final representation of every nodes.

          .. math::
              Z_{i}=\sum_{p \in P} \beta_{i, \Phi_{p}} Z_{i, \Phi_{p}}

    '''
    def __init__(self, num_edge_type, num_channels, num_layers, in_dim, hidden_dim):
        super(HSAF, self).__init__()
        self.num_channels = num_channels
        self.num_layers = num_layers
        self.in_dim = in_dim
        self.hidden_dim = hidden_dim
        self.HLHIA_layer = HLHIA(num_edge_type, self.num_channels, self.num_layers, self.in_dim, self.hidden_dim)
        # * =============== channel attention operation ================
        self.channel_attention = nn.Sequential(
            nn.Linear(self.hidden_dim, 1),
            nn.Tanh(),
            nn.Linear(1, 1, bias=False),
            nn.ReLU()
        )
        # * =============== layers attention operation ================
        self.layers_attention = nn.ModuleList()
        for i in range(num_channels):
            self.layers_attention.append(nn.Sequential(
                nn.Linear(self.hidden_dim, 1),
                nn.Tanh(),
                nn.Linear(1, 1, bias=False),
                nn.ReLU()
            ))

    def forward(self, A, h):
        attention_list = self.HLHIA_layer(A, h)
        channel_attention_list = []
        for i in range(self.num_channels):
            layer_level_feature_list = attention_list[i]
            layer_attention = self.layers_attention[i]
            for j in range(self.num_layers + 1):
                layer_level_feature = layer_level_feature_list[j]
                if j == 0:
                    layer_level_alpha = layer_attention(layer_level_feature)
                else:
                    layer_level_alpha = th.cat((layer_level_alpha, layer_attention(layer_level_feature)), dim=-1)
            layer_level_beta = th.softmax(layer_level_alpha, dim=-1)
            channel_attention_list.append(
                th.bmm(th.stack(layer_level_feature_list, dim=-1), layer_level_beta.unsqueeze(-1)).squeeze(-1))

        for i in range(self.num_channels):
            channel_level_feature = channel_attention_list[i]
            if i == 0:
                channel_level_alpha = self.channel_attention(channel_level_feature)
            else:
                channel_level_alpha = th.cat((channel_level_alpha, self.channel_attention(channel_level_feature)),
                                             dim=-1)
        channel_level_beta = th.softmax(channel_level_alpha, dim=-1)
        channel_attention = th.bmm(th.stack(channel_attention_list, dim=-1), channel_level_beta.unsqueeze(-1)).squeeze(
            -1)
        return channel_attention


class HLHIA(nn.Module):
    r"""
        Description
        -----------
        HLHIA: The Hop-Level Heterogeneous Information Aggregation

        The l-hop representation :math:`Z_{l}` is generated by the original node feature through a graph conv

        .. math::
           Z_{l}^{\Phi_{p}} = \sigma\left[\left(D_{(l)}^{\Phi_{p}}\right)^{-1} A_{(l)}^{\Phi_{p}} h W^{\Phi_{p}}\right]

        where :math:`\Phi_{p}` is the hybrid l-hop metapath and `\mathcal{h}` is the original node feature.

    """
    def __init__(self, num_edge_type, num_channels, num_layers, in_dim, hidden_dim):
        super(HLHIA, self).__init__()
        self.num_channels = num_channels
        self.in_dim = in_dim
        self.hidden_dim = hidden_dim
        layers = []
        for i in range(num_layers):
            if i == 0:
                layers.append(HMAELayer(num_edge_type, num_channels, first=True))
            else:
                layers.append(HMAELayer(num_edge_type, num_channels, first=False))
        self.layers = nn.ModuleList(layers)
        self.gcn_list = nn.ModuleList()
        for i in range(num_channels):
            self.gcn_list.append(GraphConv(in_feats=self.in_dim, out_feats=hidden_dim, norm='none', activation=F.relu))
        self.norm = EdgeWeightNorm(norm='right')

    def forward(self, A, h):
        layer_list = []
        for i in range(len(self.layers)):
            if i == 0:
                H, W, first_adj = self.layers[i](A)
                layer_list.append(first_adj)
                layer_list.append(H)
            else:
                H, W, first_adj = self.layers[i](A, H)
                layer_list.append(H)
        # * =============== GCN Encoder ================
        channel_attention_list = []
        for i in range(self.num_channels):
            gcn = self.gcn_list[i]
            layer_attention_list = []
            for j in range(len(layer_list)):
                layer = layer_list[j][i]
                layer = dgl.remove_self_loop(layer)
                edge_weight = layer.edata['w_sum']
                layer = dgl.add_self_loop(layer)
                edge_weight = th.cat((edge_weight, th.full((layer.number_of_nodes(),), 1, device=layer.device)))
                edge_weight = self.norm(layer, edge_weight)
                layer_attention_list.append(gcn(layer, h, edge_weight=edge_weight))
            channel_attention_list.append(layer_attention_list)
        return channel_attention_list


class HMAELayer(nn.Module):
    r"""
        Description
        -----------
        HMAE: Hybrid Metapath Autonomous Extraction

        The method to generate l-hop hybrid adjacency matrix

        .. math::
            A_{(l)}=\Pi_{i=1}^{l} A_{i}
    """
    def __init__(self, in_channels, out_channels, first=True):
        super(HMAELayer, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.first = first
        self.norm = EdgeWeightNorm(norm='right')
        if self.first == True:
            self.conv1 = GTConv(in_channels, out_channels, softmax_flag=False)
            self.conv2 = GTConv(in_channels, out_channels, softmax_flag=False)
        else:
            self.conv1 = GTConv(in_channels, out_channels, softmax_flag=False)

    def softmax_norm(self, H):
        norm_H = []
        for i in range(len(H)):
            g = H[i]
            g.edata['w_sum'] = self.norm(g, th.exp(g.edata['w_sum'])) # normalize the hybrid relationship matrix
            norm_H.append(g)
        return norm_H

    def forward(self, A, H_=None):
        if self.first == True:
            result_A = self.softmax_norm(self.conv1(A))
            result_B = self.softmax_norm(self.conv2(A))
            W = [self.conv1.weight.detach(), self.conv2.weight.detach()]
        else:
            result_A = H_
            result_B = self.conv1(A)
            W = [self.conv1.weight.detach().detach()]
        H = []
        for i in range(len(result_A)):
            g = dgl.adj_product_graph(result_A[i], result_B[i], 'w_sum')
            H.append(g)
        return H, W, result_A


class GTConv(nn.Module):
    r"""
        Description
        -----------

        We conv each sub adjacency matrix :math:`A_{R_{i}}` to a combination adjacency matrix :math:`A_{1}`:

        .. math::
            A_{1} = conv\left(A ; W_{c}\right)=\sum_{R_{i} \in R} w_{R_{i}} A_{R_{i}}

        where :math:`R_i \subseteq \mathcal{R}` and :math:`W_{c}` is the weight of each relation matrix
    """

    def __init__(self, in_channels, out_channels, softmax_flag=True):
        super(GTConv, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.weight = nn.Parameter(th.Tensor(out_channels, in_channels))
        self.softmax_flag = softmax_flag
        self.reset_parameters()

    def reset_parameters(self):
        nn.init.normal_(self.weight, std=0.01)

    def forward(self, A):
        if self.softmax_flag:
            Filter = F.softmax(self.weight, dim=1)
        else:
            Filter = self.weight
        num_channels = Filter.shape[0]
        results = []
        for i in range(num_channels):
            for j, g in enumerate(A):
                A[j].edata['w_sum'] = g.edata['w'] * Filter[i][j]
            sum_g = dgl.adj_sum_graph(A, 'w_sum')
            results.append(sum_g)
        return results



import argparse
import copy
import dgl
import numpy as np
import torch
import torch.nn as nn
from tqdm import tqdm
import torch.nn.functional as F
from ..models import build_model
from ..models.HeCo import LogReg
from . import BaseFlow, register_flow
from ..tasks import build_task
from ..utils import extract_embed, EarlyStopping
from sklearn.metrics import f1_score, roc_auc_score


@register_flow("HeCo_trainer")
class HeCoTrainer(BaseFlow):

    def __init__(self, args):
        super(HeCoTrainer, self).__init__(args)

        self.args = args
        self.model_name = args.model
        self.device = args.device
        self.task = build_task(args)

        self.hg = self.task.get_graph().to(self.device)
        self.num_classes = int(self.task.dataset.num_classes)
        self.args.category = self.task.dataset.category
        
        self.category = self.args.category
        self.pos = self.task.dataset.pos.to(self.device)
        self.model = build_model(self.model_name).build_model_from_args(self.args, self.hg)
        print("build_model_finish")
        self.model = self.model.to(self.device)

        self.evaluator = self.task.get_evaluator('f1')
        self.optimizer = (
            torch.optim.Adam(self.model.parameters(), lr=args.lr, weight_decay=args.weight_decay))
        self.patience = args.patience
        self.max_epoch = args.max_epoch

        self.train_idx, self.val_idx, self.test_idx = self.task.get_idx()
        self.labels = self.task.get_labels().to(self.device)

    def preprocess(self):
        super(HeCoTrainer, self).preprocess()

    def train(self):
        self.preprocess()
        stopper = EarlyStopping(self.args.patience)
        # epoch_iter = tqdm(range(self.max_epoch))
        for epoch in range(self.max_epoch):
            '''use earlyStopping'''
            loss = self._full_train_step()
            early_stop = stopper.loss_step(loss, self.model)
            print((f"Epoch: {epoch:03d}, Loss: {loss:.4f}"))
            
            if early_stop:
                print('Early Stop!\tEpoch:' + str(epoch))
                break
        
        # Evaluation
        model = stopper.load_model(self.model)
        model.eval()
        h_dict = self.model.input_feature()
        embeds = model.get_embeds(self.hg, h_dict=h_dict)
        self.evaluate(embeds,)

    def _full_train_step(self):
        self.model.train()
        self.optimizer.zero_grad()
        h_dict = self.model.input_feature()
        loss = self.model(self.hg, h_dict, self.pos)
        loss.backward()
        self.optimizer.step()
        loss = loss.cpu()
        loss = loss.detach().numpy()
        return loss

    def evaluate(self, embeds):
        hid_units = embeds.shape[1]

        xent = nn.CrossEntropyLoss()

        train_embs = embeds[self.train_idx]
        val_embs = embeds[self.val_idx]
        test_embs = embeds[self.test_idx]

        train_lbls = self.labels[self.train_idx].reshape(-1)
        val_lbls = self.labels[self.val_idx].reshape(-1)
        test_lbls = self.labels[self.test_idx].reshape(-1)
        
        accs = []
        micro_f1s = []
        macro_f1s = []
        macro_f1s_val = []
        auc_score_list = []
        for _ in range(50):
            log = LogReg(hid_units, self.num_classes)
            opt = torch.optim.Adam(log.parameters(), lr=self.args.eva_lr, weight_decay=self.args.eva_wd)
            log.to(self.device)

            val_accs = []
            test_accs = []
            val_micro_f1s = []
            test_micro_f1s = []
            val_macro_f1s = []
            test_macro_f1s = []
            
            logits_list = []
            for iter_ in range(200):
                # train
                log.train()
                opt.zero_grad()

                logits = log(train_embs)
                loss = xent(logits, train_lbls)

                loss.backward()
                opt.step()

                # val
                logits = log(val_embs)
                preds = torch.argmax(logits, dim=1)
    
                val_acc = torch.sum(preds == val_lbls).float() / val_lbls.shape[0]
                val_f1_macro = f1_score(val_lbls.cpu(), preds.cpu(), average='macro')
                val_f1_micro = f1_score(val_lbls.cpu(), preds.cpu(), average='micro')
    
                val_accs.append(val_acc.item())
                val_macro_f1s.append(val_f1_macro)
                val_micro_f1s.append(val_f1_micro)
    
                # test
                logits = log(test_embs)
                preds = torch.argmax(logits, dim=1)
    
                test_acc = torch.sum(preds == test_lbls).float() / test_lbls.shape[0]
                test_f1_macro = f1_score(test_lbls.cpu(), preds.cpu(), average='macro')
                test_f1_micro = f1_score(test_lbls.cpu(), preds.cpu(), average='micro')
    
                test_accs.append(test_acc.item())
                test_macro_f1s.append(test_f1_macro)
                test_micro_f1s.append(test_f1_micro)
                logits_list.append(logits)
    
            max_iter = val_accs.index(max(val_accs))
            accs.append(test_accs[max_iter])
            max_iter = val_macro_f1s.index(max(val_macro_f1s))
            macro_f1s.append(test_macro_f1s[max_iter])
            macro_f1s_val.append(val_macro_f1s[max_iter])
    
            max_iter = val_micro_f1s.index(max(val_micro_f1s))
            micro_f1s.append(test_micro_f1s[max_iter])
    
            # auc
            best_logits = logits_list[max_iter]
            best_proba = F.softmax(best_logits, dim=1)
            auc_score_list.append(roc_auc_score(y_true=test_lbls.detach().cpu().numpy(),
                                                y_score=best_proba.detach().cpu().numpy(),
                                                multi_class='ovr'
                                                ))
    
        
        print("\t[Classification] Macro-F1_mean: {:.4f} var: {:.4f}  Micro-F1_mean: {:.4f} var: {:.4f} auc {:.4f}"
              .format(np.mean(macro_f1s),
                      np.std(macro_f1s),
                      np.mean(micro_f1s),
                      np.std(micro_f1s),
                      np.mean(auc_score_list),
                      np.std(auc_score_list)
                      )
              )



BEST_CONFIGS = {
    "node_classification": {
        'RGCN': {
            'general': {
                'seed': 0, 'patience': 40, 'validation': True, 'evaluation_metric': 'acc'
            },
            'aifb': {
                'lr': 0.005, 'weight_decay': 0, 'max_epoch': 100,
                'hidden_dim': 32, 'n_bases': -1, 'n_layers': 2, 'batch_size': 126, 'dropout': 0,
                'mini_batch_flag': False, 'validation': True
            },
            'mutag': {
                'lr': 0.005, 'weight_decay': 0.0005, 'max_epoch': 50,
                'hidden_dim': 32, 'n_bases': 30, 'n_layers': 2, 'batch_size': 50, 'fanout': 4, 'dropout': 0.5,
                'mini_batch_flag': False, 'validation': True
            },
            'bgs': {
                'lr': 0.005, 'weight_decay': 0.0005, 'max_epoch': 50,
                'hidden_dim': 16, 'n_bases': 40, 'n_layers': 3, 'batch_size': 126, 'fanout': 4, 'dropout': 0.1,
                'mini_batch_flag': True, 'validation': True
            },
            'am': {
                'lr': 0.01, 'weight_decay': 0.0005, 'max_epoch': 50,
                'hidden_dim': 16, 'n_bases': 40, 'n_layers': 2, 'batch_size': 126, 'fanout': 4, 'dropout': 0,
                'mini_batch_flag': True, 'validation': True
            },
            'ogbn-mag': {
                'lr': 0.01, 'weight_decay': 0.0005, 'max_epoch': 10,
                'hidden_dim': 16, 'n_bases': -1, 'n_layers': 2, 'batch_size': 126, 'fanout': 4, 'dropout': 0,
                'mini_batch_flag': True, 'validation': True
            },
            'HGBn-ACM': {
                'evaluation_metric': 'f1'
            },
            'HGBn-DBLP': {
                'n_layers': 2, 'hidden_dim': 16, 'dropout': 0.5
            },
            'HGBn-Freebase': {
                'n_layers': 5, 'hidden_dim': 16, 'dropout': 0.5
            },
            'HGBn-IMDB': {
            },
            'ohgbn-Freebase': {
                'lr': 0.005, 'weight_decay': 0.0001, 'hidden_dim': 32, 'n_bases': 40, 'n_layers': 2, 'batch_size': 256,
                'fanout': 4, 'dropout': 0.2, 'mini_batch_flag': True,
            },
            'ohgbn-acm': {
                'lr': 0.01, 'hidden_dim': 64, 'dropout': 0.4, 'n_layers': 2, 'evaluation_metric': 'f1'
            },
            'ohgbn-imdb': {
                'lr': 0.01, 'hidden_dim': 128, 'dropout': 0.2, 'n_layers': 4, 'evaluation_metric': 'f1'
            },
            'ohgbn-yelp2': {'lr': 0.001, 'hidden_dim': 64, 'dropout': 0.3696343130196993, 'n_layers': 3,
                            'use_self_loop': True, 'n_bases': 3}
        },
        'CompGCN': {
            'general': {
                'seed': 0, 'patience': 100, 'validation': True, 'evaluation_metric': 'acc'
            },
            'aifb': {
                'lr': 0.01, 'weight_decay': 0.0001, 'max_epoch': 100,
                'hidden_dim': 32, 'n_layers': 2, 'dropout': 0.2, 'comp_fn': 'sub', 'batch_size': 128,
                'mini_batch_flag': False, 'validation': True
                     },
            'mutag': {
                'lr': 0.01, 'weight_decay': 0.0001, 'max_epoch': 100,
                'hidden_dim': 32, 'n_layers': 3, 'dropout': 0.5, 'comp_fn': 'sub',
                'mini_batch_flag': False, 'validation': True
            },
            'bgs': {

            },
            'am': {
                'lr': 0.001, 'weight_decay': 0.0001, 'max_epoch': 100, 'fanout': 4,
                'hidden_dim': 32, 'n_layers': 2, 'dropout': 0.2, 'comp_fn': 'sub', 'batch_size': 128,
                'mini_batch_flag': True, 'validation': True
            },
            'ohgbn-Freebase': {
                'mini_batch_flag': True, 'batch_size': 128,
            },
            'ohgbn-acm': {'lr': 0.005, 'hidden_dim': 128, 'dropout': 0.16, 'n_layers': 2,
                          'comp_fn': 'sub', 'evaluation_metric': 'f1'},
            'ohgbn-imdb': {'lr': 0.001, 'hidden_dim': 64, 'n_layers': 3,
                          'comp_fn': 'sub', 'evaluation_metric': 'f1'}
        },
        'GTN': {
            'general': {'lr': 0.005, 'weight_decay': 0.001, 'hidden_dim': 128, 'max_epoch': 100, 'patience': 20,
                        'norm_emd_flag': True, 'mini_batch_flag': False},
            'acm4GTN': {
                'num_layers': 2, 'num_channels': 2, 'adaptive_lr_flag': True,
            },
            'imdb4GTN': {
                'num_layers': 2, 'num_channels': 2, 'adaptive_lr_flag': True,
            },
            'dblp4GTN': {
                'num_layers': 3, 'num_channels': 2, 'adaptive_lr_flag': False,
            },
            'ohgbn-acm': {
                'num_layers': 2, 'num_channels': 2, 'adaptive_lr_flag': True,
            },
            'ohgbn-imdb': {
                'lr': 0.001, 'num_layers': 2, 'num_channels': 2, 'adaptive_lr_flag': True,
            }
        },
        'fastGTN': {
            'general': {'lr': 0.005, 'weight_decay': 0.001, 'hidden_dim': 64, 'max_epoch': 100, 'patience': 20,
                        'norm_emd_flag': True, 'mini_batch_flag': False},
            'acm4GTN': {
                'num_layers': 2, 'num_channels': 8, 'adaptive_lr_flag': False,
            },
            'imdb4GTN': {
                'num_layers': 2, 'num_channels': 4, 'adaptive_lr_flag': False,
            },
            'dblp4GTN': {
                'num_layers': 3, 'num_channels': 2, 'adaptive_lr_flag': False,
            },
        },
        'MHNF': {
            'general': {'lr': 0.005, 'weight_decay': 0.001, 'hidden_dim': 64, 'max_epoch': 50, 'patience': 10,
                        'norm_emd_flag': True, 'mini_batch_flag': False},
            'acm4GTN': {
                'num_layers': 2, 'num_channels': 2, 'adaptive_lr_flag': True,
            },
            'imdb4GTN': {
                'num_layers': 2, 'num_channels': 2, 'adaptive_lr_flag': True,
            },
            'dblp4GTN': {
                'num_layers': 3, 'num_channels': 2, 'adaptive_lr_flag': False,
            },
            'ohgbn-acm': {
                'lr': 0.005, 'hidden_dim': 64, 'weight_decay': 0.005, 'num_layers': 2, 'num_channels': 3,
                'norm_emd_flag': False, 'identity': False, 'adaptive_lr_flag': True,
            },
            'ohgbn-imdb': {
                'lr': 0.1, 'hidden_dim': 32, 'weight_decay': 0.01, 'num_layers': 4, 'num_channels': 2,
                'identity': False, 'adaptive_lr_flag': True,
            },
        },
        'HAN': {
            'general': {},
            'acm_han_raw': {
                'lr': 0.005, 'num_heads': [8], 'hidden_dim': 128, 'dropout': 0.6, 'weight_decay': 0.001,
                'num_epochs': 200, 'patience': 100
            },
            'acm4GTN': {
                'lr': 0.001, 'weight_decay': 0.00001, 'num_heads': [8, 8],
            },
            'imdb4GTN': {
                'lr': 0.001, 'weight_decay': 0.00001, 'num_heads': [2, 2, 2]
            },
            'HGBn-ACM': {
            },
            'ohgbn-acm': {'num_heads': [8, 8]},
            'ohgbn-imdb': {'lr': 0.005, 'hidden_dim': 32, 'dropout': 0.5}
    
        },
        'HPN': {
            'general': {},
            'ohgbn-acm': {
                'lr': 0.01, 'hidden_dim': 64, 'dropout': 0.2, 'k_layer': 3, 'alpha': 0.2, 'edge_drop': 0
          },
            'ohgbn-imdb': {
                'lr': 0.005, 'hidden_dim': 128, 'dropout': 0.25
            }
        },
        'NARS': {
            'general': {},
            'HGBn-ACM': {
                'lr': 0.001, 'num_heads': [8], 'num_hops': 3, 'hidden_dim': 64, 'dropout': 0.5, 'weight_decay': 0.001,
                'num_epochs': 200, 'patience': 100, 'evaluation_metric': 'acc', 'ff_layer': 2,
            },
            'HGBn-DBLP': {
                'lr': 0.001, 'num_heads': [8], 'num_hops': 3, 'hidden_dim': 512, 'dropout': 0.5, 'weight_decay': 0.001,
                'num_epochs': 200, 'patience': 100, 'evaluation_metric': 'acc', 'ff_layer': 2,
            },
            'HGBn-Freebase': {
                'lr': 0.001, 'num_heads': [8], 'num_hops': 2, 'hidden_dim': 64, 'dropout': 0.5, 'weight_decay': 0.001,
                'num_epochs': 200, 'patience': 100, 'evaluation_metric': 'acc', 'ff_layer': 2,
            },
            'HGBn-IMDB': {
                'lr': 0.001, 'num_heads': [8], 'num_hops': 2, 'hidden_dim': 64, 'dropout': 0.5, 'weight_decay': 0.001,
                'num_epochs': 200, 'patience': 100, 'evaluation_metric': 'acc', 'ff_layer': 2,
            },
            'ohgbn-acm': {'lr': 0.001, 'hidden_dim': 32, 'dropout': 0.4, 'num_heads': 2, 'num_hops': 2,
                          'evaluation_metric': 'f1'},
            'ohgbn-imdb': {'lr': 0.01, 'hidden_dim': 128, 'dropout': 0.4, 'num_heads': 1, 'num_hops': 2,
                           'evaluation_metric': 'f1'}


        },
        'HetGNN': {
            'general': {'max_epoch': 500, 'patience': 20, 'mini_batch_flag': True},
            'academic4HetGNN': {
                'lr': 0.001, 'weight_decay': 0.00001, 'dim': 128, 'batch_size': 64, 'window_size': 5,
                'batches_per_epoch': 50, 'rw_length': 50, 'rw_walks': 10, 'rwr_prob': 0.5,
            }
        },
        'MAGNN': {
            'general': {'mini_batch_flag': False, 'max_epoch': 500, 'patience': 30},
            'imdb': {
                'lr': 0.005, 'weight_decay': 0.001, 'dropout': 0.5, 'hidden_dim': 64, 'encoder_type': 'RotateE',
                'inter_attn_feats': 128, 'num_heads': 8, 'num_layers': 2,
            },
        },
        'NSHE': {
            'general': {},
            'acm4NSHE': {'weight_decay': 0.001, 'num_e_neg': 1, 'num_ns_neg': 4,
                         'max_epoch': 500, 'patience': 10,
                         }
        },
        'HeCo': {
            'general': {
                'hidden_dim': 64, 'max_epoch' : 10000, 'mini_batch_flag' : False
            },
            'acm4HeCo': {'seed': 84,  'eva_lr': 0.05, 'eva_wd': 0, 'patience': 5, 'learning_rate': 0.0008,
                         'weight_decay': 0, 'tau': 0.8, 'feat_drop': 0.3, 'attn_drop': 0.5,
                         'sample_rate': {'author': 7, "subject": 3}, 'lam': 0.5
            }
        },
        'DMGI': {
            'general': {
                'seed': 0,  'sc': 3, 'learning_rate' : 0.0005
            },
            'acm_han_raw': {'weight_decay': 0.0001, 'dropout': 0.5,'reg_coef': 0.001,
                            'patience': 20,'sup_coef': 0.1, 'hid_unit': 64,
                            'isSemi': False,'isBias': False,'isAttn': False
            },
            'imdb4GTN': {'weight_decay': 0.0001, 'dropout': 0.5,'reg_coef': 0.1,
                         'patience': 50,'sup_coef': 0.1, 'hid_unit': 64,
                         'isSemi': False,'isBias': False,'isAttn': False
            }
        },
        'RSHN': {
            'general': {},
            'aifb': {
                'in_dim': 16, 'hidden_dim': 16, 'num_node_layer': 2,
                'num_edge_layer': 2, 'rw_len': 4, 'batch_size': 5000,
            },
            'mutag': {
                'hidden_dim': 16, 'num_node_layer': 2, 'num_edge_layer': 2, 'rw_len': 4, 'batch_size': 5000,
                'dropout': 0.2,
            },
            'bgs': {
                'hidden_dim': 16, 'num_node_layer': 2, 'num_edge_layer': 2, 'rw_len': 4, 'batch_size': 5000
            },
            'am': {'hidden_dim': 16, 'num_node_layer': 2, 'num_edge_layer': 2, 'rw_len': 4, 'batch_size': 5000
                   },
            'ohgbn-acm': {'lr': 0.001, 'hidden_dim': 128, 'dropout': 0.4, 'num_channels': 2, 'num_node_layer': 2,
                          'num_edge_layer': 2},
            'ohgbn-imdb': {'lr': 0.01, 'hidden_dim': 128, 'dropout': 0.09, 'num_node_layer': 2, 'num_edge_layer': 2
            }
        },
        'HGSL': {
            'general': {},
            'acm4GTN': {
                'undirected_relations': 'author-paper,paper-subject', 'gnn_dropout': 0, 'fs_eps': 0.8,
                'fp_eps': 0.2, 'mp_eps': 0.6,
            }
        },
        'HGT': {
            'general': {},
            'ohgbn-acm': {
                'lr': 0.001, 'hidden_dim': 128, 'dropout': 0.1, 'n_layers': 2, 'num_heads': 1
            },
            'ohgbn-imdb': {'lr': 0.001, 'hidden_dim': 128, 'dropout': 0.3, 'n_layers': 2, 'num_heads': 4},
    
        },
        'SimpleHGN': {
            'general': {},
            'ohgbn-acm': {
                'lr': 0.0005, 'hidden_dim': 64, 'num_layers': 3, 'edge_dim': 32, 'num_heads': 2, 'num_edge': 4,
                'attn_drop_rate': 0.1, 'feats_drop_rate': 0.5, 'beta': 0.05, 'residual': False, 'slope': 0.01,
                'max_epoch': 200, 'patience': 50
            },
            'ohgbn-imdb': {'lr': 0.001, 'hidden_dim': 32, 'num_layers': 2, 'edge_dim': 32, 'num_heads': 8, 'num_edge': 8,
                'attn_drop_rate': 0.34, 'feats_drop_rate': 0.4, 'beta': 0.01,
                'residual': False, 'slope': 0.1, 'max_epoch': 100, 'patience': 50},

        }

    },
    
    "link_prediction": {
        'NARS': {
            'general': {'num_hops': 3},
        },
        'HetGNN': {
            'general': {'max_epoch': 500, 'patience': 10, 'mini_batch_flag': True},
            'academic4HetGNN': {
                'lr': 0.01, 'weight_decay': 0.0001, 'dim': 128, 'batch_size': 64, 'window_size': 5,
                'batches_per_epoch': 50, 'rw_length': 50, 'rw_walks': 10, 'rwr_prob': 0.5,
            }
        },
        'HAN': {
            'general':{
            
            },
            'HGBl-PubMed': {
                'lr': 0.01, 'weight_decay': 0.0005, 'max_epoch': 400,
                'hidden_dim': 64, 'n_bases': 40, 'n_layers': 4, 'batch_size': 126, 'fanout': 4, 'dropout': 0.5,
                'validation': True, 'feat': 2, 'seed': 4,
            },
        },
        'CompGCN':{
            'general': {
            },
            'ohgbl-yelp1': {
                'mini_batch_flag': True, 'n_layers': 2,
            },
            'ohgbl-yelp2': {
                'mini_batch_flag': True, 'n_layers': 2, 'batch_size': 1024
            },
            'ohgbl-MTWM': {
                'weight_decay': 0.00001, 'max_epoch': 50, 'patience': 50,
                'mini_batch_flag': True, 'n_layers': 2,
            }
        },
        'RGCN': {
            'general': {
            },
            'HGBl-amazon': {
                'lr': 0.01, 'weight_decay': 0.0005, 'max_epoch': 400,
                'hidden_dim': 64, 'n_bases': 40, 'n_layers': 4, 'batch_size': 126, 'fanout': 4, 'dropout': 0.5,
                'validation': True, 'feat': 2, 'seed': 0,
            },
            'HGBl-PubMed': {
                'lr': 0.01, 'weight_decay': 0.0005, 'max_epoch': 400,
                'hidden_dim': 64, 'n_bases': 40, 'n_layers': 4, 'batch_size': 126, 'fanout': 4, 'dropout': 0.5,
                'validation': True, 'feat': 2, 'seed': 4,
            },
            'HGBl-LastFM': {
                'lr': 0.01, 'weight_decay': 0.0005, 'max_epoch': 400,
                'hidden_dim': 64, 'n_bases': 40, 'n_layers': 4, 'batch_size': 126, 'fanout': 4, 'dropout': 0.5,
                'validation': True, 'feat': 2, 'seed': 3,
            },
            'FB15k-237': {
                'lr': 0.01, 'weight_decay': 0.0005, 'max_epoch': 100,
                'hidden_dim': 16, 'n_bases': 40, 'n_layers': 2, 'batch_size': 126, 'fanout': 4, 'dropout': 0,
                'validation': True
            },
            'ohgbl-MTWM': {
                'mini_batch_flag': True, 'n_layers': 2, 'batch_size': 12800
            },
            'ohgbl-yelp1': {
                'mini_batch_flag': True, 'n_layers': 2, 'batch_size': 102400
            },
            'ohgbl-yelp2': {
                'mini_batch_flag': True, 'n_layers': 2,
            }
        },

    },
    "recommendation": {
        'KGCN': {
            "general": {},
            'LastFM4KGCN': {
                'in_dim': 16, 'hidden_dim': 16, 'n_relation': 60,
                'batch_size': 128, 'lr' : 0.002
            },
        }
    },
}


import torch
import torch.nn as nn
import torch.nn.functional as F
from ..layers.GeneralHGNNLayer import HeteroGeneralLayer
from ..layers.GeneralGNNLayer import GeneralLayer


def HGNNLayer(gnn_type, rel_names, dim_in, dim_out, dropout, act, has_bn, has_l2norm, **kwargs):
    return HeteroGeneralLayer(gnn_type, rel_names, dim_in, dim_out, dropout, act, has_bn, has_l2norm, **kwargs)


class HGNNSkipBlock(nn.Module):
    '''Skip block for HGNN'''

    def __init__(self, gnn_type, rel_names, dim_in, dim_out, num_layers,
                 stage_type, dropout, act, has_bn, has_l2norm, **kwargs):
        super(HGNNSkipBlock, self).__init__()
        self.stage_type = stage_type
        self.f = nn.ModuleList()
        if num_layers == 1:
            self.f.append(HGNNLayer(gnn_type, rel_names, dim_in, dim_out, dropout, act, has_bn, has_l2norm, **kwargs))
        else:
            self.f = []
            for i in range(num_layers - 1):
                d_in = dim_in if i == 0 else dim_out
                self.f.append(HGNNLayer(gnn_type, rel_names, d_in, dim_out, dropout, act, has_bn, has_l2norm, **kwargs))
            d_in = dim_in if num_layers == 1 else dim_out
            self.f.append(HGNNLayer(gnn_type, rel_names, d_in, dim_out, dropout, act, has_bn, has_l2norm, **kwargs))
        self.act = act
        if stage_type == 'skipsum':
            assert dim_in == dim_out, 'Sum skip must have same dim_in, dim_out'

    def forward(self, g, h):
        h_0 = h
        for layer in self.f:
            h = layer(g, h)
        out_h = {}
        for key, value in h_0.items():

            if self.stage_type == 'skipsum':
                out_h[key] = self.act(h[key] + h_0[key])
            elif self.stage_type == 'skipconcat':
                out_h[key] = self.act(torch.cat((h[key], h_0[key]), 1))
            else:
                raise ValueError('stage_type must in [skipsum, skipconcat]')

        return out_h


########### Stage: NN except start and head ############

class HGNNStackStage(nn.Module):
    '''Simple Stage that stack GNN layers'''

    def __init__(self, gnn_type, rel_names, stage_type, dim_in, dim_out,
                 num_layers, skip_every, dropout, act, has_bn, has_l2norm, **kwargs):
        super(HGNNStackStage, self).__init__()

        for i in range(num_layers):
            d_in = dim_in if i == 0 else dim_out
            layer = HGNNLayer(gnn_type, rel_names, d_in, dim_out, dropout, act, has_bn, has_l2norm, **kwargs)
            self.add_module('layer{}'.format(i), layer)
        self.dim_out = dim_out
        self.has_l2norm = has_l2norm

    def forward(self, g, h_dict):
        for layer in self.children():
            h_dict = layer(g, h_dict)
        if self.has_l2norm:
            for name, batch_h in h_dict.items():
                h_dict[name] = F.normalize(batch_h, p=2, dim=-1)
        return h_dict


class HGNNSkipStage(nn.Module):
    ''' Stage with skip connections'''

    def __init__(self, gnn_type, rel_names, stage_type, dim_in, dim_out,
                 num_layers, skip_every, dropout, act, has_bn, has_l2norm, **kwargs):
        super(HGNNSkipStage, self).__init__()
        assert num_layers % skip_every == 0, \
            'cfg.gnn.skip_every must be multiples of cfg.gnn.layer_mp' \
            '(excluding head layer)'
        for i in range(num_layers // skip_every):
            if stage_type == 'skipsum':
                d_in = dim_in if i == 0 else dim_out
            elif stage_type == 'skipconcat':
                d_in = dim_in if i == 0 else dim_in + i * dim_out
            block = HGNNSkipBlock(gnn_type, rel_names, d_in, dim_out, skip_every, stage_type, dropout, act, has_bn,
                                  has_l2norm, **kwargs)
            self.add_module('block{}'.format(i), block)
        if stage_type == 'skipconcat':
            self.dim_out = d_in + dim_out
        else:
            self.dim_out = dim_out
        self.has_l2norm = has_l2norm

    def forward(self, g, h_dict):
        for layer in self.children():
            h_dict = layer(g, h_dict)
        if self.has_l2norm:
            for name, batch_h in h_dict.items():
                h_dict[name] = F.normalize(batch_h, p=2, dim=-1)
        return h_dict


def GNNLayer(gnn_type, dim_in, dim_out, dropout, act, has_bn, has_l2norm, **kwargs):
    return GeneralLayer(gnn_type, dim_in, dim_out, dropout, act, has_bn, has_l2norm, **kwargs)


class GNNSkipBlock(nn.Module):
    r"""
    Skip block for HGNN
    """

    def __init__(self, gnn_type, dim_in, dim_out, num_layers, stage_type, dropout, act, has_bn, has_l2norm, *args, **kwargs):
        super(GNNSkipBlock, self).__init__()
        self.stage_type = stage_type
        self.f = nn.ModuleList()
        if num_layers == 1:
            self.f.append(GNNLayer(gnn_type, dim_in,
                                   dim_out, dropout, None, has_bn, has_l2norm, **kwargs))
        else:
            for i in range(num_layers - 1):
                d_in = dim_in if i == 0 else dim_out
                self.f.append(GNNLayer(gnn_type, d_in, dim_out, dropout, act, has_bn, has_l2norm, **kwargs))
            d_in = dim_in if num_layers == 1 else dim_out
            self.f.append(GNNLayer(gnn_type, d_in, dim_out, dropout, None, has_bn, has_l2norm, **kwargs))
        self.act = act
        if stage_type == 'skipsum':
            assert dim_in == dim_out, 'Sum skip must have same dim_in, dim_out'

    def forward(self, g, h):
        h_0 = h
        for layer in self.f:
            h = layer(g, h)
        if self.stage_type == 'skipsum':
            h = h + h_0
        elif self.stage_type == 'skipconcat':
            h = torch.cat((h_0, h), 1)
        else:
            raise ValueError('stage_type must in [skipsum, skipconcat]')
        h = self.act(h)
        return h


########### Stage: NN except start and head ############

class GNNStackStage(nn.Module):
    '''Simple Stage that stack GNN layers'''

    def __init__(self, gnn_type, dim_in, dim_out,
                 num_layers, dropout, act, has_bn, has_l2norm, *args, **kwargs):
        super(GNNStackStage, self).__init__()

        for i in range(num_layers):
            d_in = dim_in if i == 0 else dim_out
            # if i == num_layers - 1:
            #     layer = GNNLayer(gnn_type, d_in, dim_out, dropout, None, has_bn, has_l2norm, **kwargs)
            # else:
            layer = GNNLayer(gnn_type, d_in, dim_out, dropout, act, has_bn, has_l2norm, **kwargs)
            self.add_module('layer{}'.format(i), layer)
        self.dim_out = dim_out
        self.has_l2norm = has_l2norm

    def forward(self, g, h):
        for layer in self.children():
            h = layer(g, h)
        if self.has_l2norm:
            h = F.normalize(h, p=2, dim=-1)
        return h


class GNNSkipStage(nn.Module):
    ''' Stage with skip connections'''

    def __init__(self, gnn_type, stage_type,dim_in, dim_out, num_layers,
                 skip_every, dropout, act, has_bn, has_l2norm, *args, **kwargs):
        super(GNNSkipStage, self).__init__()
        assert num_layers % skip_every == 0, \
            'cfg.gnn.skip_every must be multiples of cfg.gnn.layer_mp' \
            '(excluding head layer)'
        for i in range(num_layers // skip_every):
            if stage_type == 'skipsum':
                d_in = dim_in if i == 0 else dim_out
            elif stage_type == 'skipconcat':
                d_in = dim_in if i == 0 else dim_in + i * dim_out
            # if i == num_layers // skip_every - 1:
            #     block = GNNSkipBlock(gnn_type, d_in, dim_out, skip_every, stage_type, dropout, None, has_bn, has_l2norm,
            #                          **kwargs)
            # else:
            block = GNNSkipBlock(gnn_type, d_in, dim_out, skip_every, stage_type, dropout, act, has_bn, has_l2norm, **kwargs)
            self.add_module('block{}'.format(i), block)
        if stage_type == 'skipconcat':
            self.dim_out = d_in + dim_out
        else:
            self.dim_out = dim_out
        self.has_l2norm = has_l2norm

    def forward(self, g, h):
        for layer in self.children():
            h = layer(g, h)
        if self.has_l2norm:
            h = F.normalize(h, p=2, dim=-1)
        return h


import torch as th
from dgl.dataloading.negative_sampler import _BaseNegativeSampler, Uniform
from dgl import backend as F


class NegativeSampler(object):
    def __init__(self, g, k):
        # caches the probability distribution
        self.weights = {
            etype: g.in_degrees(etype=etype).float() ** 0.75
            for etype in g.etypes
        }
        self.k = k

    def __call__(self, g, eids_dict):
        result_dict = {}
        for etype, eids in eids_dict.items():
            src, _ = g.find_edges(eids, etype=etype)
            src = src.repeat_interleave(self.k)
            dst = self.weights[etype].multinomial(len(src), replacement=True)
            result_dict[etype] = (src, dst)
        return result_dict


class Multinomial(_BaseNegativeSampler):
    """Negative sampler that randomly chooses negative destination nodes
    for each source node according to a uniform distribution.

    For each edge ``(u, v)`` of type ``(srctype, etype, dsttype)``, DGL generates
    :attr:`k` pairs of negative edges ``(u, v')``, where ``v'`` is chosen
    uniformly from all the nodes of type ``dsttype``.  The resulting edges will
    also have type ``(srctype, etype, dsttype)``.

    Parameters
    ----------
    k : int
        The number of negative examples per edge.

    Examples
    --------
    >>> g = dgl.graph(([0, 1, 2], [1, 2, 3]))
    >>> neg_sampler = dgl.dataloading.negative_sampler.Uniform(2)
    >>> neg_sampler(g, [0, 1])
    (tensor([0, 0, 1, 1]), tensor([1, 0, 2, 3]))
    """
    def __init__(self, g, k):
        self.weights = {
            etype: g.in_degrees(etype=etype).float() ** 0.75
            for etype in g.canonical_etype
        }
        self.k = k

    def _generate(self, g, eids, canonical_etype):
        src, _ = g.find_edges(eids, etype=canonical_etype)
        src = F.repeat(src, self.k, 0)
        dst = self.weights[canonical_etype].multinomial(len(src), replacement=True)
        return src, dst


class Uniform_exclusive(_BaseNegativeSampler):
    """Negative sampler that randomly chooses negative destination nodes
    for each source node according to a uniform distribution.

    For each edge ``(u, v)`` of type ``(srctype, etype, dsttype)``, DGL generates
    :attr:`k` pairs of negative edges ``(u, v')``, where ``v'`` is chosen
    uniformly from all the nodes of type ``dsttype``.  The resulting edges will
    also have type ``(srctype, etype, dsttype)``.

    Parameters
    ----------
    k : int
        The number of negative examples per edge.

    Examples
    --------
    >>> g = dgl.graph(([0, 1, 2], [1, 2, 3]))
    >>> neg_sampler = dgl.dataloading.negative_sampler.Uniform(2)
    >>> neg_sampler(g, [0, 1])
    (tensor([0, 0, 1, 1]), tensor([1, 0, 2, 3]))
    """
    def __init__(self, k):
        self.k = k

    def _generate(self, g, eids, canonical_etype):
        src, _ = g.find_edges(eids, etype=canonical_etype)
        src_list = []
        dst_list = []
        for i in range(len(eids)):
            s = src[i]
            exp = g.successors(s, etype=canonical_etype)
            dst = th.randint(low=0, high=g.number_of_nodes(canonical_etype[2]), size=(self.k,))
            for d in range(len(dst)):
                while dst[d] in exp:
                    dst[d] = th.randint(low=0, high=g.number_of_nodes(canonical_etype[2]), size=(1,))
            s = s.repeat_interleave(self.k)
            src_list.append(s)
            dst_list.append(dst)
        return th.cat(src_list), th.cat(dst_list)


import logging
import os
import colorlog
import re
import datetime
from logging import getLogger
from colorama import init


def printInfo(metric, epoch, train_score, train_loss, val_score, val_loss):
    if metric == 'f1_lr':
        print((
            f"Epoch: {epoch:03d}, Train_loss: {train_loss:.4f}, Train_macro_f1: {train_score[0]:.4f}, Train_micro_f1: {train_score[1]:.4f}, "
            f"Val_macro_f1: {val_score[0]:.4f}, Val_micro_f1: {val_score[1]:.4f}, ValLoss:{val_loss: .4f}"
        ))
    # use acc
    elif metric == 'acc':
        print((
            f"Epoch: {epoch:03d}, Train_loss: {train_loss:.4f}, Train_acc: {train_score:.4f},  "
            f"Val_acc: {val_score:.4f}, ValLoss:{val_loss: .4f}"
        ))
    elif metric == 'acc-ogbn-mag':
        print((
            f"Epoch: {epoch:03d}, Train_loss: {train_loss:.4f}, Train_acc: {train_score:.4f},  "
            f"Val_acc: {val_score:.4f}, ValLoss:{val_loss: .4f}"
        ))
    else:
        print((
            f"Epoch: {epoch:03d}, Train_loss: {train_loss:.4f}, Train_macro_f1: {train_score[0]:.4f}, Train_micro_f1: {train_score[1]:.4f}, "
            f"Val_macro_f1: {val_score[0]:.4f}, Val_micro_f1: {val_score[1]:.4f}, ValLoss:{val_loss: .4f}"
        ))


def printMetric(metric, score, mode):
    if isinstance(score, tuple):
        print(f"{mode}_macro_{metric} = {score[0]:.4f}, {mode}_micro_{metric}: {score[1]:.4f}")
    else:
        print(f"{mode}_{metric} = {score:.4f}")


# UPDATE
# Hu Anke 2021/11/07

log_colors_config = {
    'DEBUG': 'cyan',
    'WARNING': 'yellow',
    'ERROR': 'red',
    'CRITICAL': 'red',
}


def get_local_time():
    r"""Get current time

    Returns:
        str: current time
    """
    cur = datetime.datetime.now()
    cur = cur.strftime('%b-%d-%Y_%H-%M-%S')

    return cur


def ensure_dir(dir_path):
    r"""Make sure the directory exists, if it does not exist, create it

    Args:
        dir_path (str): directory path

    """
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)


class RemoveColorFilter(logging.Filter):

    def filter(self, record):
        if record:
            ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
            record.msg = ansi_escape.sub('', str(record.msg))
        return True


def set_color(log, color, highlight=True):
    color_set = ['black', 'red', 'green', 'yellow', 'blue', 'pink', 'cyan', 'white']
    try:
        index = color_set.index(color)
    except:
        index = len(color_set) - 1
    prev_log = '\033['
    if highlight:
        prev_log += '1;3'
    else:
        prev_log += '0;3'
    prev_log += str(index) + 'm'
    return prev_log + log + '\033[0m'


# UPDATE
# Hu Anke 2021/11/11
# UPDATE
# Hu AnKe 2021/12/16
class Logger:

    def __init__(self, config):
        """
        A logger that can show a message on standard output and write it into the
        file named `filename` simultaneously.
        All the message that you want to log MUST be str.

        Args:
            config (Config): An instance object of Config, used to record parameter information.

        Example:
            >>> logger = logging.getLogger(config)
            >>> logger.debug(train_state)
            >>> logger.info(train_result)
        """
        init(autoreset=True)
        LOGROOT = f'./openhgnn/output/{config.model}/'
        dir_name = os.path.dirname(LOGROOT)
        ensure_dir(dir_name)

        logfilename = '{}-{}.log'.format(config.model, get_local_time())

        logfilepath = os.path.join(LOGROOT, logfilename)

        filefmt = "%(asctime)-15s %(levelname)s  %(message)s"
        filedatefmt = "%a %d %b %Y %H:%M:%S"
        fileformatter = logging.Formatter(filefmt, filedatefmt)

        sfmt = "%(log_color)s%(asctime)-15s %(levelname)s  %(message)s"
        sdatefmt = "%d %b %H:%M"
        sformatter = colorlog.ColoredFormatter(sfmt, sdatefmt, log_colors=log_colors_config)
        if not hasattr(config, 'state') or config.state.lower() == 'info':
            level = logging.INFO
        elif config.state.lower() == 'debug':
            level = logging.DEBUG
        elif config.state.lower() == 'error':
            level = logging.ERROR
        elif config.state.lower() == 'warning':
            level = logging.WARNING
        elif config.state.lower() == 'critical':
            level = logging.CRITICAL
        else:
            level = logging.INFO

        fh = logging.FileHandler(logfilepath, mode='a')
        fh.setLevel(level)
        fh.setFormatter(fileformatter)
        remove_color_filter = RemoveColorFilter()
        fh.addFilter(remove_color_filter)

        sh = logging.StreamHandler()
        sh.setLevel(level)
        sh.setFormatter(sformatter)
        
        root_logger = logging.getLogger()
        for h in root_logger.handlers:
            root_logger.removeHandler(h)

        logging.basicConfig(level=level, handlers=[sh, fh])
        self.logger = getLogger()
        self.logger.info(config)

    def info(self, s):
        self.logger.info(s)
        
    def load_best_config(self, s):
        self.logger.info('[Load Best Config] ' + s)
        
    def train_info(self, s):
        self.logger.info('[Train Info] ' + s)
    
    def metric2str(self, metric_dict):
        out = "[Evaluation metric]"
        for mode, score_dict in metric_dict.items():
            out += f"\tMode:{mode}, "
            for metric, score in score_dict.items():
                out += f"{metric}: {score:.4f}; "
        return out
    
    def dataset_info(self, s):
        self.logger.info('[Dataset Process] ' + s)
        
    def feature_info(self, s):
        self.logger.info('[Feature Transformation] ' + s)
    
    # graph data analyze
    def log_data_info(self, g):
        num_nodes = g.num_nodes()
        num_edges = g.num_edges()
        node_types = len(g.ntypes)
        edge_types = len(g.etypes)
        c_etypes = len(g.canonical_etypes)
        datainfo = {'total nodes':num_nodes, 'total edges':num_edges, 'node types':node_types, 'edge types': edge_types,
                    'c_etypes': c_etypes}

        self.logger.info(datainfo)
        return

    # evaluate results
    def log_eval_info(self, metric, epoch, train_score, train_loss, val_score, val_loss):
        if metric == 'f1_lr':
            eval_info = {f"Epoch: {epoch:03d}, Train_loss: {train_loss:.4f}, Train_macro_f1: {train_score[0]:.4f}, Train_micro_f1: {train_score[1]:.4f}, "
                f"Val_macro_f1: {val_score[0]:.4f}, Val_micro_f1: {val_score[1]:.4f}, ValLoss:{val_loss: .4f}"}

        # use acc
        elif metric == 'acc':
            eval_info = {f"Epoch: {epoch:03d}, Train_loss: {train_loss:.4f}, Train_acc: {train_score:.4f},  "
                f"Val_acc: {val_score:.4f}, ValLoss:{val_loss: .4f}"}

        elif metric == 'acc-ogbn-mag':
            eval_info = {f"Epoch: {epoch:03d}, Train_loss: {train_loss:.4f}, Train_acc: {train_score:.4f},  "
                f"Val_acc: {val_score:.4f}, ValLoss:{val_loss: .4f}"}

        else:
            eval_info = {f"Epoch: {epoch:03d}, Train_loss: {train_loss:.4f}, Train_macro_f1: {train_score[0]:.4f}, Train_micro_f1: {train_score[1]:.4f}, "
                f"Val_macro_f1: {val_score[0]:.4f}, Val_micro_f1: {val_score[1]:.4f}, ValLoss:{val_loss: .4f}"}
        self.logger.info(eval_info)
        return

    def log_metric_info_1(self, metric, score, mode):
        met_info = {f"{mode}_{metric} = {score:.4f}"}
        self.logger.info(met_info)
        return

    def log_metric_info_2(self, metric, score, mode):
        met_info = {f"{mode}_macro_{metric} = {score[0]:.4f}, {mode}_micro_{metric}: {score[1]:.4f}"}
        self.logger.info(met_info)


import dgl
import dgl.function as fn
import numba
import numpy as np
import torch as th
import scipy.sparse as ssp
import array
import torch



class Budget(object):
    def __init__(self, hg, n_types, NS):
        self.n_types = {}
        for key, value in n_types.items():
            self.n_types[key] = th.zeros(value)
        self.NS = NS
        self.hg = hg
    def update(self, dst_type, idxs):
        for etype in self.hg.canonical_etypes:
            if dst_type == etype[2]:
                src_type = etype[0]
                #degree = self.hg.in_degrees(idx, etype=etype)
                for i in idxs:
                    src_idx = self.hg.predecessors(i, etype=etype)
                    #if src_idx.shape[0] > 0:
                    len = src_idx.shape[0]
                    if src_type in self.NS.keys():
                        src_idx = th.tensor([i for i in src_idx if i not in self.NS[src_type]])
                    if src_idx.shape[0] > 0:
                        self.n_types[src_type][src_idx] += 1 / len

    def pop(self, type, idx):
        self.n_types[type][idx] = 0


class HGTsampler(object):
    def __init__(self, hg, category, num_nodes_per_type, num_steps):
        self.n_types = {}
        for n in hg.ntypes:
            self.n_types[n] = hg.num_nodes(n)
        self.category = category
        self.num_nodes_per_type = num_nodes_per_type
        self.num_steps = num_steps
        self.hg = hg

    def sampler_subgraph(self, seed_nodes):
        OS = {self.category: th.stack(seed_nodes)}
        NS = OS
        B = Budget(self.hg, self.n_types, NS)
        for type, idxs in OS.items():
            B.update(type, idxs)
        for i in range(self.num_steps):
            prob = {}
            for src_type, p in B.n_types.items():
                #print(src_type)
                if p.max() > 0:
                    prob[src_type] = p / th.sum(p)
                    sampled_idx = th.multinomial(prob[src_type], self.num_nodes_per_type, replacement=False)
                    if not OS.__contains__(src_type):
                        OS[src_type] = sampled_idx
                    else:
                        OS[src_type] = th.cat((OS[src_type], sampled_idx))
                    B.update(src_type, sampled_idx)
                    B.pop(src_type, sampled_idx)
        sg = self.hg.subgraph(OS)
        return sg, OS




def HGT_preprocess4mag(hg, train_idx):
    hg = hg.to('cpu')
    edges = {etype: hg.edges(etype=etype) for etype in hg.canonical_etypes}
    edges.update({(v, e + '_inv', u): (dst, src) for (u, e, v), (src, dst) in edges.items()})
    hg2 = dgl.heterograph(edges)
    hg2 = dgl.to_simple(hg2)

    # Initialize year
    hg2.nodes['paper'].data['timestamp'] = hg.nodes['paper'].data['year'].squeeze()
    for ntype in hg.ntypes:
        if ntype != 'paper':
            hg2.nodes[ntype].data['timestamp'] = th.zeros(hg2.num_nodes(ntype), dtype=th.int64)

    # Aggregate bag-of-paper features
    hg2.nodes['paper'].data['feat'] = hg.nodes['paper'].data['feat']
    hg2.update_all(fn.copy_u('feat', 'm'), fn.mean('m', 'feat'), etype='has_topic')  # field_of_study
    hg2.update_all(fn.copy_u('feat', 'm'), fn.mean('m', 'feat'), etype='writes_inv')  # author
    hg2.update_all(fn.copy_u('feat', 'm'), fn.mean('m', 'feat'), etype='affiliated_with')  # institution

    # Attach log-degree to feature of each node type
    for ntype in hg2.ntypes:
        hg2.nodes[ntype].data['deg'] = th.zeros(hg2.num_nodes(ntype))
    for utype, etype, vtype in hg2.canonical_etypes:
        hg2.nodes[vtype].data['deg'] += hg2.in_degrees(etype=etype)
    for ntype in hg2.ntypes:
        hg2.nodes[ntype].data['feat'] = th.cat([
            hg2.nodes[ntype].data['feat'],
            th.log10(hg2.nodes[ntype].data['deg'][:, None])], 1)
        del hg2.nodes[ntype].data['deg']

    for ntype in hg2.ntypes:
        hg2.nodes[ntype].data['train_mask'] = torch.zeros(hg2.num_nodes(ntype), dtype=torch.bool)
        if ntype == 'paper':
            hg2.nodes[ntype].data['train_mask'][train_idx['paper']] = True

    # Convert to homogeneous graph and add self-loop
    g = dgl.to_homogeneous(hg2, ndata=['timestamp', 'feat'])
    g.edata['etype'] = g.edata[dgl.ETYPE]
    g.ndata['ntype'] = g.ndata[dgl.NTYPE]
    g.ndata['nid'] = g.ndata[dgl.NID]
    del g.edata[dgl.ETYPE]
    del g.edata[dgl.EID]
    del g.ndata[dgl.NTYPE]
    del g.ndata[dgl.NID]
    num_nodes = g.num_nodes()
    g = dgl.add_self_loop(g)
    g.edata['etype'][-num_nodes:] = len(hg2.etypes)

    return g


import dgl
import time
import tqdm
import torch as th
import numpy as np
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
from openhgnn.utils.sampler import get_epoch_samples
from openhgnn.utils.dgl_graph import give_one_hot_feats, normalize_edges, edata_in_out_mask, load_link_pred
from openhgnn.utils.utils import print_dict, h2dict
from openhgnn.utils.evaluator import evaluate, cal_loss_f1


def cal_node_pairwise_loss(node_emd, edge, neg_edge):
    # cross entropy loss from LINE
    # pos loss
    inner_product = cal_inner_product(node_emd, edge)
    pos_loss = - th.mean(F.logsigmoid(inner_product))
    # neg loss
    inner_product = cal_inner_product(node_emd, neg_edge)
    neg_loss = - th.mean(F.logsigmoid(-1 * inner_product))
    loss = pos_loss + neg_loss
    return loss

def cal_inner_product(node_emd, edge):
     emb_u_i = node_emd[edge[0]]
     emb_u_j = node_emd[edge[1]]
     inner_product = th.sum(emb_u_i * emb_u_j, dim=1)
     return inner_product

def cal_cla_loss(predict, ns_label):
    BCE_loss = th.nn.BCELoss()
    return BCE_loss(predict, ns_label)

def run(model, g, config):
    g_homo = dgl.to_homogeneous(g)
    pos_edges = g_homo.edges()
    optimizer = optim.Adam(model.parameters(), lr=config.lr)
    evaluate(config.seed, config.dataset, g.ndata['h'], g)
    model.train()
    for epoch in range(config.max_epoch):
        epoch_start_time = time.time()
        neg_edges, ns_samples = get_epoch_samples(g, epoch, config.dataset, config.num_ns_neg, config.device)

        optimizer.zero_grad()
        node_emb, ns_prediction, eva_h = model(g, ns_samples)
        # compute loss
        pairwise_loss = cal_node_pairwise_loss(node_emb, pos_edges, neg_edges)

        ns_label = th.cat([ns['label'] for ns in ns_samples]).type(th.float32).to(config.device)
        cla_loss = cal_cla_loss(ns_prediction, ns_label)
        loss = pairwise_loss + cla_loss * config.beta
        loss.backward()
        epoch_dict = {'Epoch': epoch, 'train_loss': loss.item(),
                      'pairwise_loss': pairwise_loss.item(),
                      'cla_loss': cla_loss.item(),
                      'time': time.time() - epoch_start_time}
        print_dict(epoch_dict, '\n')
        optimizer.step()
    model.eval()
    node_emb, ns_prediction, eva_h = model(g, ns_samples)
    evaluate(config.seed, config.dataset, eva_h, g)
    return eva_h




def run_GTN(model, g, config):
    if config.adaptive_lr == 'False':
        optimizer = th.optim.Adam(model.parameters(), lr=config.lr, weight_decay=config.weight_decay)
    else:
        optimizer = th.optim.Adam([{'params': model.weight},
                                      {'params': model.linear1.parameters()},
                                      {'params': model.linear2.parameters()},
                                      {"params": model.layers.parameters(), "lr": 0.5}
                                      ], lr=0.005, weight_decay=0.001)

    best_val_loss = 10000
    best_test_loss = 10000
    best_train_loss = 10000
    best_train_f1 = 0
    best_val_f1 = 0
    best_test_f1 = 0
    loss_func = th.nn.CrossEntropyLoss()
    g_homo = dgl.to_homogeneous(g, ndata=['h'])
    for i in range(config.max_epoch):
        for param_group in optimizer.param_groups:
            if param_group['lr'] > 0.005:
                param_group['lr'] = param_group['lr'] * 0.9
        print('Epoch:  ', i + 1)
        model.zero_grad()
        model.train()

        y_train = model(g_homo)
        emd = h2dict(y_train, g.ndata['h'])
        tar_y = emd['paper']
        tar_data = g.nodes['paper'].data

        #evaluate(config.seed, config.dataset, g.ndata['h'], g)
        # evaluate(config.seed, config.dataset, h2dict(X_, g.ndata['h']), g)
        # evaluate(config.seed, config.dataset, h2dict(a, g.ndata['h']), g)

        loss, macro_f1, micro_f1 = cal_loss_f1(tar_y, tar_data, loss_func, 'train_mask')
        print('Train - Loss: {:.4f}, Macro_F1: {:.4f}, Micro_F1: {:.4f}'.format(loss.item(), macro_f1, micro_f1))
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        train_f1 = macro_f1
        # Valid
        model.eval()
        with th.no_grad():
            #y_test = models.forward(g_homo)
            emd = h2dict(y_train, g.ndata['h'])
            tar_y = emd['paper']
            loss, macro_f1, micro_f1 = cal_loss_f1(tar_y, tar_data, loss_func, 'valid_mask')
            print('Valid - Loss: {:.4f}, Macro_F1: {:.4f}, Micro_F1: {:.4f}'.format(loss.item(), macro_f1, micro_f1))
            val_f1 = macro_f1

            loss, macro_f1, micro_f1 = cal_loss_f1(tar_y, tar_data, loss_func, 'test_mask')
            print('Test - Loss: {:.4f}, Macro_F1: {:.4f}, Micro_F1: {:.4f}'.format(loss.item(), macro_f1, micro_f1))
            test_f1 = macro_f1
            if val_f1 > best_val_f1:
                best_train_f1 = train_f1
                best_val_f1 = val_f1
                best_test_f1 = test_f1
        #th.empty_cache()
    print('---------------Best Results--------------------')
    print('Train -  Macro_F1: {}'.format(best_train_f1))
    print('Valid -   Macro_F1: {}'.format(best_val_f1))
    print('Test -  Macro_F1: {}'.format(best_test_f1))


def get_idx(hg, g, category):
    train_mask = hg.nodes[category].data.pop('train_mask')
    test_mask = hg.nodes[category].data.pop('test_mask')
    train_idx = th.nonzero(train_mask, as_tuple=False).squeeze()
    test_idx = th.nonzero(test_mask, as_tuple=False).squeeze()
    labels = hg.nodes[category].data.pop('labels')
    # get target category id
    node_ids = th.arange(g.number_of_nodes())
    category_id = len(hg.ntypes)
    for i, ntype in enumerate(hg.ntypes):
        if ntype == category:
            category_id = i
    # find out the target node ids in g
    node_tids = g.ndata[dgl.NTYPE]
    loc = (node_tids == category_id)
    target_idx = node_ids[loc]
    return target_idx, train_idx, test_idx, labels


def run_RSHN(model, hg, cl_graph, config):

    optimizer = th.optim.Adam(model.parameters(), lr=config.lr, weight_decay=config.weight_decay)
    g = dgl.to_homogeneous(hg)
    target_idx, train_idx, test_idx, labels = get_idx(hg, g, config.category)
    if config.validation == True:
        val_idx = train_idx[:len(train_idx) // 5]
        train_idx = train_idx[len(train_idx) // 5:]
    best_test_acc = 0
    for epoch in range(config.max_epoch):
        print('Epoch:  ', epoch + 1)
        model.train()
        optimizer.zero_grad()
        logits = model(g, cl_graph)

        logits = logits[target_idx]
        loss = F.cross_entropy(logits[train_idx], labels[train_idx])

        loss.backward()
        optimizer.step()

        train_acc = th.sum(logits[train_idx].argmax(dim=1) == labels[train_idx]).item() / len(train_idx)
        if config.validation == True:
            val_loss = F.cross_entropy(logits[val_idx], labels[val_idx])
            val_acc = th.sum(logits[val_idx].argmax(dim=1) == labels[val_idx]).item() / len(val_idx)
            print("Train Accuracy: {:.4f} | Train Loss: {:.4f} | Validation Accuracy: {:.4f} | Validation loss: {:.4f}".
                  format(train_acc, loss.item(), val_acc, val_loss.item()))
        else:
            print("Train Accuracy: {:.4f} | Train Loss: {:.4f} ".
              format(train_acc, loss.item()))

        model.eval()
        logits = model.forward(g, cl_graph)
        logits = logits[target_idx]
        test_loss = F.cross_entropy(logits[test_idx], labels[test_idx])
        test_acc = th.sum(logits[test_idx].argmax(dim=1) == labels[test_idx]).item() / len(test_idx)
        print("Test Accuracy: {:.4f} | Test loss: {:.4f}".format(test_acc, test_loss.item()))
        print()
        if test_acc > best_test_acc:
            best_test_acc = test_acc

    print('Test - ACC: {}'.format(best_test_acc))


    return


def run_CompGCN(model, hg, config):

    # optimizer
    optimizer = th.optim.Adam(model.parameters(), lr=config.lr, weight_decay=config.weight_decay)

    # training loop
    print("start training...")
    forward_time = []
    backward_time = []

    edata_in_out_mask(hg)
    g = dgl.to_homogeneous(hg, edata=['in_edges_mask', 'out_edges_mask'])

    target_idx, train_idx, test_idx, labels = get_idx(hg, g, config.category)
    if config.validation:
        val_idx = train_idx[:len(train_idx) // 5]
        train_idx = train_idx[len(train_idx) // 5:]
    else:
        val_idx = train_idx
    model.train()
    best_test_acc = 0
    for epoch in range(config.max_epoch):
        optimizer.zero_grad()
        t0 = time.time()
        model.train()
        logits = model(g)
        logits = logits[target_idx]
        loss = F.cross_entropy(logits[train_idx], labels[train_idx])
        t1 = time.time()
        loss.backward()
        optimizer.step()
        t2 = time.time()

        forward_time.append(t1 - t0)
        backward_time.append(t2 - t1)
        print("Epoch {:05d} | Train Forward Time(s) {:.4f} | Backward Time(s) {:.4f}".
              format(epoch, forward_time[-1], backward_time[-1]))
        train_acc = th.sum(logits[train_idx].argmax(dim=1) == labels[train_idx]).item() / len(train_idx)
        val_loss = F.cross_entropy(logits[val_idx], labels[val_idx])
        val_acc = th.sum(logits[val_idx].argmax(dim=1) == labels[val_idx]).item() / len(val_idx)
        print("Train Accuracy: {:.4f} | Train Loss: {:.4f} | Validation Accuracy: {:.4f} | Validation loss: {:.4f}".
              format(train_acc, loss.item(), val_acc, val_loss.item()))
        print()
        model.eval()
        logits = model.forward(g)
        logits = logits[target_idx]
        test_loss = F.cross_entropy(logits[test_idx], labels[test_idx])
        test_acc = th.sum(logits[test_idx].argmax(dim=1) == labels[test_idx]).item() / len(test_idx)
        print("Test Accuracy: {:.4f} | Test loss: {:.4f}".format(test_acc, test_loss.item()))
        print()
        print()
        if test_acc > best_test_acc:
            best_test_acc = test_acc

        # print("Mean forward time: {:4f}".format(np.mean(forward_time[len(forward_time) // 4:])))
        # print("Mean backward time: {:4f}".format(np.mean(backward_time[len(backward_time) // 4:])))
    print('Test - ACC: {}'.format(best_test_acc))


def run_RGCN(model, hg, config):

    # optimizer
    optimizer = th.optim.Adam(model.parameters(), lr=config.lr, weight_decay=config.l2norm)

    # training loop
    print("start training...")
    forward_time = []
    backward_time = []

    # calculate norm for each edge type and store in edge
    hg = normalize_edges(hg, 'norm')
    g = dgl.to_homogeneous(hg, edata=['norm'])

    # since the nodes are featureless, the input feature is then the node id.
    feats = th.arange(g.number_of_nodes()).to(config.device)

    edge_type = g.edata[dgl.ETYPE].long()
    edge_norm = g.edata['norm']

    target_idx, train_idx, test_idx, labels = get_idx(hg, g, config.category)
    if config.validation:
        val_idx = train_idx[:len(train_idx) // 5]
        train_idx = train_idx[len(train_idx) // 5:]
    else:
        val_idx = train_idx
    model.train()
    for epoch in range(config.max_epoch):
        optimizer.zero_grad()
        t0 = time.time()
        logits = model(g, feats)
        logits = logits[target_idx]
        loss = F.cross_entropy(logits[train_idx], labels[train_idx])
        t1 = time.time()
        loss.backward()
        optimizer.step()
        t2 = time.time()

        forward_time.append(t1 - t0)
        backward_time.append(t2 - t1)
        print("Epoch {:05d} | Train Forward Time(s) {:.4f} | Backward Time(s) {:.4f}".
              format(epoch, forward_time[-1], backward_time[-1]))
        train_acc = th.sum(logits[train_idx].argmax(dim=1) == labels[train_idx]).item() / len(train_idx)
        val_loss = F.cross_entropy(logits[val_idx], labels[val_idx])
        val_acc = th.sum(logits[val_idx].argmax(dim=1) == labels[val_idx]).item() / len(val_idx)
        print("Train Accuracy: {:.4f} | Train Loss: {:.4f} | Validation Accuracy: {:.4f} | Validation loss: {:.4f}".
              format(train_acc, loss.item(), val_acc, val_loss.item()))
        print()
        model.eval()
        logits = model.forward(g, feats, edge_type, edge_norm)
        logits = logits[target_idx]
        test_loss = F.cross_entropy(logits[test_idx], labels[test_idx])
        test_acc = th.sum(logits[test_idx].argmax(dim=1) == labels[test_idx]).item() / len(test_idx)
        print("Test Accuracy: {:.4f} | Test loss: {:.4f}".format(test_acc, test_loss.item()))
        print()

        print("Mean forward time: {:4f}".format(np.mean(forward_time[len(forward_time) // 4:])))
        print("Mean backward time: {:4f}".format(np.mean(backward_time[len(backward_time) // 4:])))



class NegativeSampler(object):
    def __init__(self, g, k):
        # caches the probability distribution
        self.weights = {
            etype: g.in_degrees(etype=etype).float() ** 0.75
            for _, etype, _ in g.canonical_etypes
        }
        self.k = k

    def __call__(self, g, eids_dict):
        result_dict = {}
        for etype, eids in eids_dict.items():
            src, _ = g.find_edges(eids, etype=etype)
            src = src.repeat_interleave(self.k)
            dst = self.weights[etype].multinomial(len(src), replacement=True)
            result_dict[etype] = (src, dst)
        return result_dict




def run_HetGNN(model, hg, het_graph, config):
    # het_graph is used to sample neighbour
    hg = hg.to('cpu')
    category = config.category
    train_mask = hg.nodes[category].data.pop('train_mask')
    test_mask = hg.nodes[category].data.pop('test_mask')
    train_idx = th.nonzero(train_mask, as_tuple=False).squeeze()
    test_idx = th.nonzero(test_mask, as_tuple=False).squeeze()
    labels = hg.nodes[category].data.pop('label')
    label_mask = hg.nodes[category].data.pop('label_mask')
    emd =hg.nodes[category].data['dw_embedding']
    train_batch = load_link_pred('./openhgnn/dataset/a_a_list_train.txt')
    test_batch = load_link_pred('./openhgnn/dataset/a_a_list_test.txt')
    author_link_prediction(emd, train_batch, test_batch)
    micro_f1, macro_f1 = nc_with_split(emd, labels, train_idx, test_idx)
    print('<Classification> DW    Micro-F1 = %.4f, Macro-F1 = %.4f' % (micro_f1, macro_f1))
    # HetGNN Sampler
    from torch.utils.data import IterableDataset, DataLoader
    from openhgnn.utils.sampler import SkipGramBatchSampler, HetGNNCollator, NeighborSampler
    batch_sampler = SkipGramBatchSampler(hg, config.batch_size, config.window_size)
    neighbor_sampler = NeighborSampler(het_graph, hg.ntypes, batch_sampler.num_nodes, config.device)
    collator = HetGNNCollator(neighbor_sampler, hg)
    dataloader = DataLoader(
        batch_sampler,
        collate_fn=collator.collate_train,
        num_workers=config.num_workers)

    # sampler = dgl.dataloading.MultiLayerFullNeighborSampler(1)
    # train_eid_dict = {etype: hg.edges(etype=etype, form='eid')
    #     for etype in hg.etypes}
    opt = th.optim.Adam(model.parameters())
    # dataloader = dgl.dataloading.EdgeDataLoader(
    #     # The following arguments are specific to NodeDataLoader.
    #     hg,  # The graph
    #     train_eid_dict,  # The edges to iterate over
    #     sampler,  # The neighbor sampler
    #     negative_sampler=NegativeSampler(hg, 1),  # The negative sampler
    #     device=config.device,  # Put the MFGs on CPU or GPU
    #     # The following arguments are inherited from PyTorch DataLoader.
    #     batch_size=1024,  # Batch size
    #     shuffle=True,  # Whether to shuffle the nodes for every epoch
    #     drop_last=False,  # Whether to drop the last incomplete batch
    #     num_workers=0  # Number of sampler processes
    # )
    pred = ScorePredictor()
    dataloader_it = iter(dataloader)
    het_graph = het_graph.to(config.device)
    input_features = extract_feature(het_graph, hg.ntypes)
    x = model(het_graph, input_features)
    author_link_prediction(x['author'].to('cpu').detach(), train_batch, test_batch)

    micro_f1, macro_f1 = nc_with_split(x[config.category].to('cpu').detach(), labels, train_idx, test_idx)
    print('<Classification>     Micro-F1 = %.4f, Macro-F1 = %.4f' % (micro_f1, macro_f1))
    for i in range(config.max_epoch):
        model.train()
        for batch_id in tqdm.trange(config.batches_per_epoch):
            positive_graph, negative_graph, blocks = next(dataloader_it)
            blocks = [b.to(config.device) for b in blocks]
            positive_graph = positive_graph.to(config.device)
            negative_graph = negative_graph.to(config.device)
            # we need extract multi-feature
            input_features = extract_feature(blocks[0], hg.ntypes)

            x = model(blocks[0], input_features)
            loss = compute_loss(pred(positive_graph, x), pred(negative_graph, x))

            opt.zero_grad()
            loss.backward()
            opt.step()
        print('Epoch {:05d} |Train - Loss: {:.4f}'.format(i, loss.item()))
        input_features = extract_feature(het_graph, hg.ntypes)
        x = model(het_graph, input_features)
        author_link_prediction(x['author'].to('cpu').detach(), train_batch, test_batch)
        micro_f1, macro_f1 = nc_with_split(x[config.category].to('cpu').detach(), labels, train_idx, test_idx)
        print('<Classification>     Micro-F1 = %.4f, Macro-F1 = %.4f' % (micro_f1, macro_f1))
    pass


class ScorePredictor(nn.Module):
    def forward(self, edge_subgraph, x):
        with edge_subgraph.local_scope():
            edge_subgraph.ndata['x'] = x
            for etype in edge_subgraph.canonical_etypes:
                edge_subgraph.apply_edges(
                    dgl.function.u_dot_v('x', 'x', 'score'), etype=etype)
            return edge_subgraph.edata['score']

def compute_loss(pos_score, neg_score):
    # an example hinge loss
    loss = []
    for i in pos_score:
        loss.append(F.logsigmoid(pos_score[i]))
        loss.append(F.logsigmoid(-neg_score[i]))
    loss = th.cat(loss)
    return -loss.mean()


def extract_feature(g, ntypes):
    input_features = {}
    for n in ntypes:
        ndata = g.srcnodes[n].data
        data = {}
        data['dw_embedding'] = ndata['dw_embedding']
        data['abstract'] = ndata['abstract']
        if n == 'paper':
            data['title'] = ndata['title']
            data['venue'] = ndata['venue']
            data['author'] = ndata['author']
            data['reference'] = ndata['reference']
        input_features[n] = data

    return input_features



import yaml
from random import choice
import argparse
import os


hidden_dim = [8, 16, 32, 64, 128]
layers_pre_mp = [1, 2, 3]
layers_post_mp = [1, 2, 3]
layers_gnn = [1, 2, 3, 4, 5, 6]
stage_type = ['stack', 'skipsum', 'skipconcat']
activation = ['relu', 'lrelu', 'elu', 'sigmoid', 'tanh',]
has_bn = [True, False]
has_l2norm = [True, False]
mini_batch_flag = False
macro_func = ['attention', 'sum', 'mean', 'max']
dropout = [0.0, 0.3, 0.6]
lr = [0.1, 0.01, 0.001, 0.0001]
weight_decay = 0.0001
patience = 40
max_epoch = [100, 200, 400]
optimizer = ['Adam', 'SGD']
num_heads = [1, 2, 4, 8]
featn = [0, 1, 2]
featl = [0, 2]
loss_fn = ['distmult', 'dot-product']
gnn_type = ['gcnconv', 'gatconv', 'ginconv', 'sageconv']


def makeDict(gnn_type, type):
    dict = {
        'hidden_dim': choice(hidden_dim),
        'layers_pre_mp': choice(layers_pre_mp),
        'layers_post_mp': choice(layers_post_mp),
        'layers_gnn': choice(layers_gnn),
        'stage_type': choice(stage_type),
        'activation': choice(activation),
        'dropout': choice(dropout),
        'has_bn': choice(has_bn),
        'feat': choice(featn) if type == 'node' else choice(featl),
        'has_l2norm': choice(has_l2norm),
        'lr': choice(lr),
        'weight_decay': weight_decay,
        'patience': patience,
        'max_epoch': choice(max_epoch),
        'mini_batch_flag': mini_batch_flag,
        'macro_func': choice(macro_func),
        'optimizer': choice(optimizer),
        'num_heads': choice(num_heads),
        'loss_fn': choice(loss_fn) if type == 'link' else None,
        'gnn_type': gnn_type,
    }
    return dict

def generate(gnn_type, i, key, configfile):
    datasets_node = ['HGBn-ACM', 'HGBn-IMDB', 'HGBn-DBLP', 'HGBn-Freebase', 'HNE-PubMed']
    datasets_link = ['HGBl-amazon', 'HGBl-LastFM', 'HGBl-PubMed', 'HGBl-ACM', 'HGBl-DBLP', 'HGBl-IMDB']
    datasets_rec = ['yelp4HeGAN', 'DoubanMovie']
    dicts = {}
    dicts2 = {}
    dicts3 = {}

    for a in datasets_node:
        dict = makeDict(gnn_type, 'node')
        dicts[a] = dict
    for a in datasets_link:
        dict = makeDict(gnn_type, 'link')
        dicts2[a] = dict
    for a in datasets_rec:
        dict = makeDict(gnn_type, 'rec')
        dicts3[a] = dict

    aproject = {'node_classification': dicts,
                'link_prediction': dicts2,
                'recommendation': dicts3
    }

    fileNamePath = os.path.split(os.path.realpath(__file__))[0]
    path = fileNamePath + '/config/{}'.format(configfile)
    if not os.path.exists(path):
        os.makedirs(path)
    path = '{}/{}'.format(path, key)
    if not os.path.exists(path):
        os.makedirs(path)
    name = gnn_type + '_' + i + '.yaml'
    yamlPath = os.path.join(path, name)

    with open(yamlPath, 'w') as f:
        yaml.dump(aproject, f)
        print('Generate yaml file successfully!')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--gnn_type', '-a', default='gcnconv', type=str, help='gnn type')
    parser.add_argument('--times', '-s', default='1', type=str, help='times')
    parser.add_argument('--key', '-k', default='has_bn', type=str, help='attribute')
    parser.add_argument('--configfile', '-c', default='test', type=str, help='config file path')

    args = parser.parse_args()
    generate(args.gnn_type, args.times, args.key, args.configfile)



import copy
import dgl
import torch as th
import numpy as np
from tqdm import tqdm
import torch.nn.functional as F
from torch.utils.data import DataLoader
from ..models import build_model
from . import BaseFlow, register_flow
from ..tasks import build_task
from ..sampler.HetGNN_sampler import SkipGramBatchSampler, HetGNNCollator, NeighborSampler, hetgnn_graph
from ..utils import EarlyStopping


@register_flow("deepwalktrainer")
class DeepwalkTrainer(BaseFlow):
    """DeepwalkTrainer flows.

    Supported Model: DeepWalk
    Supported Dataset：Academic4HetGNN
    Dataset description can be found in HetGNN paper.
    The trainerflow supports node classification and author link prediction.

    """

    def __init__(self, args):
        super(DeepwalkTrainer, self).__init__(args)

        self.args = args
        self.dataset = DeepwalkDataset(
            net_file=args.data_file,
            map_file=args.map_file,
            walk_length=args.walk_length,
            window_size=args.window_size,
            num_walks=args.num_walks,
            batch_size=args.batch_size,
            negative=args.negative,
            gpus=args.gpus,
            fast_neg=args.fast_neg,
            ogbl_name=args.ogbl_name,
            load_from_ogbl=args.load_from_ogbl,
        )
        self.emb_size = self.dataset.G.number_of_nodes()
        self.emb_model = None

    def preprocess(self):

        if self.args.mini_batch_flag:
            if self.args.model == 'HetGNN':
                hetg = hetgnn_graph(self.hg, self.args.dataset)
                self.hg = self.hg.to('cpu')
                self.het_graph = hetg.get_hetgnn_graph(self.args.rw_length, self.args.rw_walks, self.args.rwr_prob).to('cpu')

                batch_sampler = SkipGramBatchSampler(self.hg, self.args.batch_size, self.args.window_size)
                neighbor_sampler = NeighborSampler(self.het_graph, self.hg.ntypes, batch_sampler.num_nodes, self.args.device)
                collator = HetGNNCollator(neighbor_sampler, self.hg)
                dataloader = DataLoader(
                    batch_sampler,
                    collate_fn=collator.collate_train,
                    num_workers=self.args.num_workers)
                self.dataloader_it = iter(dataloader)
                self.hg = self.hg.to(self.args.device)
                self.het_graph = self.het_graph.to(self.args.device)
            # elif self.args.model == 'Metapath2vec':
            #     batch_sampler = SkipGramBatchSampler(self.hg, self.args.batch_size, self.args.window_size, self.args.rw_length)
            #     collator = MP2vecCollator(self.hg.ntypes, batch_sampler.num_nodes)
            #     dataloader = DataLoader(batch_sampler, collate_fn=collator.collate_train, num_workers=self.args.num_workers)
            #     self.dataloader_it = iter(dataloader)

        return

    def train(self):
        self.preprocess()
        stopper = EarlyStopping(self.args.patience, self._checkpoint)
        epoch_iter = tqdm(range(self.max_epoch))
        for epoch in epoch_iter:
            if self.args.mini_batch_flag:
                loss = self._mini_train_step()
            else:
                loss = self._full_train_setp()
            epoch_iter.set_description('Epoch{}: Loss:{:.4f}'.format(epoch, loss))
            early_stop = stopper.loss_step(loss, self.model)
            if early_stop:
                print('Early Stop!\tEpoch:' + str(epoch))
                break
        stopper.load_model(self.model)
        metrics = self._test_step()
        return dict(metrics=metrics)

    def _full_train_setp(self):
        self.model.train()
        negative_graph = self.construct_negative_graph()
        x = self.model(self.het_graph)[self.category]
        loss = self.loss_calculation(self.ScorePredictor(self.hg, x), self.ScorePredictor(negative_graph, x))
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()
        return loss.item()

    def init_device_emb(self):
        """ set the device before training
        will be called once in fast_train_mp / fast_train
        """
        choices = sum([self.args.only_gpu, self.args.only_cpu, self.args.mix])
        assert choices == 1, "Must choose only *one* training mode in [only_cpu, only_gpu, mix]"

        # initializing embedding on CPU
        self.emb_model = SkipGramModel(
            emb_size=self.emb_size,
            emb_dimension=self.args.dim,
            walk_length=self.args.walk_length,
            window_size=self.args.window_size,
            batch_size=self.args.batch_size,
            only_cpu=self.args.only_cpu,
            only_gpu=self.args.only_gpu,
            mix=self.args.mix,
            neg_weight=self.args.neg_weight,
            negative=self.args.negative,
            lr=self.args.lr,
            lap_norm=self.args.lap_norm,
            fast_neg=self.args.fast_neg,
            record_loss=self.args.print_loss,
            norm=self.args.norm,
            use_context_weight=self.args.use_context_weight,
            async_update=self.args.async_update,
            num_threads=self.args.num_threads,
        )

        torch.set_num_threads(self.args.num_threads)
        if self.args.only_gpu:
            print("Run in 1 GPU")
            assert self.args.gpus[0] >= 0
            self.emb_model.all_to_device(self.args.gpus[0])
        elif self.args.mix:
            print("Mix CPU with %d GPU" % len(self.args.gpus))
            if len(self.args.gpus) == 1:
                assert self.args.gpus[0] >= 0, 'mix CPU with GPU should have available GPU'
                self.emb_model.set_device(self.args.gpus[0])
        else:
            print("Run in CPU process")
            self.args.gpus = [torch.device('cpu')]

    def fast_train(self):
        """ fast train with dataloader with only gpu / only cpu"""
        # the number of postive node pairs of a node sequence
        num_pos = 2 * self.args.walk_length * self.args.window_size \
                  - self.args.window_size * (self.args.window_size + 1)
        num_pos = int(num_pos)

        self.init_device_emb()

        if self.args.async_update:
            self.emb_model.share_memory()
            self.emb_model.create_async_update()

        sampler = self.dataset.create_sampler(0)

        dataloader = DataLoader(
            dataset=sampler.seeds,
            batch_size=self.args.batch_size,
            collate_fn=sampler.sample,
            shuffle=False,
            drop_last=False,
            num_workers=self.args.num_sampler_threads,
        )

        num_batches = len(dataloader)
        print("num batchs: %d\n" % num_batches)

        start_all = time.time()
        start = time.time()
        with torch.no_grad():
            max_i = num_batches
            for i, walks in enumerate(dataloader):
                if self.args.fast_neg:
                    self.emb_model.fast_learn(walks)
                else:
                    # do negative sampling
                    bs = len(walks)
                    neg_nodes = torch.LongTensor(
                        np.random.choice(self.dataset.neg_table,
                                         bs * num_pos * self.args.negative,
                                         replace=True))
                    self.emb_model.fast_learn(walks, neg_nodes=neg_nodes)

                if i > 0 and i % self.args.print_interval == 0:
                    if self.args.print_loss:
                        print("Batch %d training time: %.2fs loss: %.4f" \
                              % (i, time.time() - start, -sum(self.emb_model.loss) / self.args.print_interval))
                        self.emb_model.loss = []
                    else:
                        print("Batch %d, training time: %.2fs" % (i, time.time() - start))
                    start = time.time()

            if self.args.async_update:
                self.emb_model.finish_async_update()

        print("Training used time: %.2fs" % (time.time() - start_all))
        if self.args.save_in_txt:
            self.emb_model.save_embedding_txt(self.dataset, self.args.output_emb_file)
        elif self.args.save_in_pt:
            self.emb_model.save_embedding_pt(self.dataset, self.args.output_emb_file)
        else:
            self.emb_model.save_embedding(self.dataset, self.args.output_emb_file)



